﻿//-------[ COLLISION MANAGER ]---------------------------------------------------
/* Purpose: Collision Manager global manager that handles all collision checking
in game.  Divides the screen into various bins and compartments, to which game
objects pop in and out.  Every frame look for collisions between various types
of things in a given bin (players vs enemies, etc.) and notify game objects of
collisions by calling their collide() method.
*/
//-------------------------------------------------------------------------------
"use scrict";


class CollisionManager {

    //-------[ CONSTRUCTOR ]----------------------------------------------------
    /*  Purpose: base constructor for CollisionManager
    */
    //--------------------------------------------------------------------------
    constructor() {
        //pattern to ensure singleton of GAME_WINDOW object
        //there can be only one...
        if (!AACOLLISION_MANAGER) {
            this._secretProp = 4;

            AACOLLISION_MANAGER = this;
        } else {
            console.log("AAENGINE: ERROR: Attempt to create multiple collision managers." +
                        "\nAAENGINE: ERROR: Only first instanciation will be used.");
            return AACOLLISION_MANAGER;
        }
    }
    //--------------------------------------------------------------------------


    //-------[ INIT ]-----------------------------------------------------------
    /*  Purpose: initialize the collision manager with appropriate values and 
    setup routines.  NOTE: init must be called after the game window is 
    constructed and initialized as it uses values from game window in init
    */
    //--------------------------------------------------------------------------
    init(bX, bY) {
        if (!this.initialized) {

            //arrays of bins and associated tracking vars
            this._collisionBins = [];
            this._numBinsX = bX;
            this._numBinsY = bY;
            this._halfNumBinsX = Math.round(this._numBinsX / 2);
            this._halfNumBinsY = Math.round(this._numBinsY / 2);

            //width and height of a given bin
            this._binWidth = GAME_WINDOW.GAME_WINDOW_UNSCALED_WIDTH / bX;
            this._binHeight = GAME_WINDOW.GAME_WINDOW_UNSCALED_HEIGHT / bY;

            //arrays for points to be used over and over
            this._pointA = [0, 0];
            this._pointB = [0, 0];
            this._pointC = [0, 0];
            this._pointD = [0, 0];

            for (var i = 0; i < bX; i++) { //bins in X
                this._collisionBins[i] = [];
                for (var j = 0; j < bY; j++) { //bins in Y
                    this._collisionBins[i][j] = new Bin(
                        i * this._binWidth, //left
                        j * this._binHeight, //top
                        (i + 1) * this._binWidth, //right
                        (j + 1) * this._binHeight, //bottom
                        this); //reference to the manager

                }
            }

            this.initialized = true;
        }
    }
    //--------------------------------------------------------------------------


    //-------[ PREUPDATE ]------------------------------------------------------
    /*  Purpose: called each frame BEFORE game objects have been updated, 
    clears all bins of existing data and gets ready for the frame.
    */
    //--------------------------------------------------------------------------
    preupdate() {
        for (var i = 0; i < this._numBinsX; i++) {
            for (var j = 0; j < this._numBinsY; j++) {
                var b = this._collisionBins[i][j];
                b._gOs[0].length = 0; //clear the array of children
                b._gOs[1].length = 0; //clear the array of children
                b._gOs[2].length = 0; //clear the array of children
                b._gOs[3].length = 0; //clear the array of children
            }
        }
    }
    //--------------------------------------------------------------------------


    //-------[ UPDATE ]---------------------------------------------------------
    /*  Purpose: called each frame AFTER game objects have been updated, checks
    for collisions between various elements in each bin.  (Objects update their
    bin as a part of their update cycle).  Any collisions call the 'collide'
    method of a given game object with a reference to whatever hit them.
    */
    //--------------------------------------------------------------------------
    update() {

        //update all bins, pop objects on and off...
        for (var i = 0; i < this._numBinsX; i++) {
            for (var j = 0; j < this._numBinsY; j++) {
                var b = this._collisionBins[i][j];
                
                //0 = player, 1=playerBullet, 2=enemy, 3=enemyBullet

                //loop through all players and check against
                //things in bin.  usually only 1 or 2 players
                for (var k = 0; k < b._gOs[0].length; k++) {
                    var player = b._gOs[0][k];

                    //check current player against all enemies
                    for (var l = 0; l < b._gOs[2].length; l++) {
                        var result = this.checkForCollision(
                            player, b._gOs[2][l])

                        if (result) {
                            //if we collided, call collide method
                            //on both player and enemy
                            player.collide(b._gOs[2][l]);
                            b._gOs[2][l].collide(player);
                            
                        }
                    }

                    //check current player against all enemies bullets
                    for (var l = 0; l < b._gOs[3].length; l++) {
                        var result = this.checkForCollision(
                            player, b._gOs[3][l])

                        if (result) {
                            //if we collided, call collide method
                            //on both player and enemy
                            player.collide(b._gOs[3][l]);
                            b._gOs[4][l].collide(player);

                        }
                    }
                }
               
                
                //loop through all enemies and check against
                //player bullets
                for (var k = 0; k < b._gOs[2].length; k++) {
                    var enemy = b._gOs[2][k];

                    //check current player against all enemies
                    for (var l = 0; l < b._gOs[1].length; l++) {
                        var result = this.checkForCollision(
                            enemy, b._gOs[1][l])

                        if (result) {
                            //if we collided, call collide method
                            //on both enemy and player bullet
                            enemy.collide(b._gOs[1][l]);
                            b._gOs[1][l].collide(enemy);

                        }
                    }
                }
                

                //if we are debugging and not threading, draw 
                //all the bins we just processed
                if(DEBUG && !USE_THREADING) { this.drawBin(b) }
            }
        }
    }
    //--------------------------------------------------------------------------


    //-------[ DRAW BIN ]-------------------------------------------------------
    /*  Purpose: draws a bin on the debug overlay.  Should only be called if
    operating in /non threading/ mode, as threaded mode doesnt have access to 
    the main UI thread / canvas.  Drawing of bins is handled elsewhere in such
    cases.
    */
    //--------------------------------------------------------------------------
    drawBin(bin) {
        var x = Math.round((bin._x1 / GAME_WINDOW.GAME_WINDOW_UNSCALED_WIDTH)*255);
        var y = Math.round((bin._y1 / GAME_WINDOW.GAME_WINDOW_UNSCALED_HEIGHT)*255);

        if (bin._gOs[0].length > 0) {
            DEBUG_DISPLAY.drawDebugBox(
                bin._x1 * GAME_WINDOW.GAME_WINDOW_SCALE,
                bin._y1 * GAME_WINDOW.GAME_WINDOW_SCALE,
                this._binWidth * GAME_WINDOW.GAME_WINDOW_SCALE,
                this._binHeight * GAME_WINDOW.GAME_WINDOW_SCALE,
                'rgba(' + x + ',0,' + y + ',0.5)',
                'rgba(255,255,255,0.5)');
        } else {
            DEBUG_DISPLAY.drawDebugBox(
                bin._x1 * GAME_WINDOW.GAME_WINDOW_SCALE,
                bin._y1 * GAME_WINDOW.GAME_WINDOW_SCALE,
                this._binWidth * GAME_WINDOW.GAME_WINDOW_SCALE,
                this._binHeight * GAME_WINDOW.GAME_WINDOW_SCALE,
                'rgba(' + x + ',0,' + y + ',0.1)',
                'rgba(255,255,255,0.5)'
            );
        }
    }
    //--------------------------------------------------------------------------


    //-------[ ADD OBJECT TO BINS ]---------------------------------------------
    /*  Purpose: adds an object to whatever set of bins it should be in based 
    on screen location.  
    */
    //--------------------------------------------------------------------------
    addObjectToBins(GO) {
        var bins = this.addShapeToBins(GO._adjusted_points, GO.POS2D);
        for (var i = 0; i < bins.length; i++) {
            this._collisionBins[bins[i][0]][bins[i][1]]._gOs[GO._typeID].push(GO);
        }
    }
    //--------------------------------------------------------------------------


    //-------[ ADD SHAPE TO BINS ]----------------------------------------------
    /* Purpose: adds a shape (collection of points) to bins, returning an array
    of all bins that the points land inside of.  NOTE: does not accound for 
    shapes that eclipse a bin but whose points are all outside.  Used by most
    other 'add X funcions' when they are adding things to bins for collisions
    */
    //--------------------------------------------------------------------------
    addShapeToBins(points, loc) {

        var currentBins = [];
        for (var i = 0; i < points.length; i++) {
           
            var binX = Math.floor(((loc[0] + points[i][0]) /
                this._binWidth) + (this._numBinsX * 0.5));
            var binY = Math.floor((((-loc[1] + points[i][1])) /
                this._binHeight) + (this._numBinsY * 0.5));

            if (binX > -1 && binX < this._numBinsX) {
                if (binY > -1 && binY < this._numBinsY) {
                    var tempBin = [binX, binY];

                    var isInCurBins = false;
                    for (var j = 0; j < currentBins.length; j++) {
                        var a = currentBins[j][0];
                        var b = currentBins[j][1];
                        if (a == binX && b == binY) {
                            isInCurBins = true;
                        }
                    }

                    if (!isInCurBins) {
                        currentBins.push(tempBin);
                    }
                }
            }
        }
        return currentBins;
    }
    //--------------------------------------------------------------------------


    //-------[ checkForCollision ]----------------------------------------------
    /* Purpose: checks for collision between two game objects.  Called from
    inside the collision manager, in a loop that checks various elements of 
    a bin against other elements in a bin (players vs. enemy bullets, enemies
    vs palyer bullets, etc.)  Handles various permutations of different types
    of game objects (cube, circle, poly)
    */
    //--------------------------------------------------------------------------
    checkForCollision(gO, gO2) {
        //CASE 1 - BOX / BOX COLLISION DETECTION
        //NOTE - DOES NOT CONSIDER "CONTAINMENT" AS COLLISION
        if (gO._type == 0 && gO2._type == 0) {

            var x1 = gO._2dPosX + gO._adjusted_points[0][0];
            var y1 = gO._2dPosY + gO._adjusted_points[0][1];
            var x2 = gO._2dPosX + gO._adjusted_points[2][0];
            var y2 = gO._2dPosY + gO._adjusted_points[2][1];

            for (var i = 0; i < 4; i++) {

                var x = gO2._2dPosX + gO2._adjusted_points[i][0];
                var y = gO2._2dPosY + gO2._adjusted_points[i][1];
                if (x >= x1 && x <= x2) {
                    if (y >= y1 && y <= y2) {
                        return true;
                    }
                }
            }
        } //end box/box case

        //CASE 2 - CIRCLE / CIRCLE COLLISION DETECTION
        if (gO._type == 1 && gO2._type == 1) {
            var dX = gO2._2dPosX - gO._2dPosX;
            var dY = gO2._2dPosY - gO._2dPosY;
            var radii = gO._radius + gO2._radius;

            if (((dX * dX) + (dY * dY)) <= (radii * radii)) {
                return true;
            }
        } //end circle/circle case
        
        //CASE 3 - CIRCLE / SQUARE (order matters)
        if (gO._type == 1 && gO2._type == 0) {
            return this.circleSquareCollisionCheck(gO, gO2);
        }

        if (gO._type == 0 && gO2._type == 1) {
            return this.circleSquareCollisionCheck(gO2, gO);
        }

        //CASE 4 - POLY / POLY
        if (gO._type == 2 && gO2._type == 2) {
            return this.polyPolyCollisionCheck(gO, gO2);
        }

        //CASE 5 - POLY / BOX (either order)
        if (gO._type == 2 && gO2._type == 0) {
            return this.polyPolyCollisionCheck(gO, gO2);
        }
        if (gO._type == 0 && gO2._type == 2) {
            return this.polyPolyCollisionCheck(gO, gO2);
        }
        //CASE 6 - POLY / CIRCLE
        if (gO._type == 1 && gO2._type == 2) {
            return this.circlePolyCollisionCheck(gO, gO2);
        }

        if (gO._type == 2 && gO2._type == 1) {
            return this.circlePolyCollisionCheck(gO2, gO);
        }
    }
    //--------------------------------------------------------------------------

    //-------[ polyPolyCollisionCheck ]-----------------------------------------
    /* Purpose: checks for collision between two game objects that are both 
    marked as n-gon convex polygon collision shapes.
    */
    //--------------------------------------------------------------------------
    polyPolyCollisionCheck(gO, gO2) {
        //console.log("possible poly collision " + Date.now());
        var bIntersects = false;
        var s = GAME_WINDOW.GAME_WINDOW_SCALE;
        for (var i = 0; i < gO._adjusted_points.length; i++) {
            
            //get line segment A - B from first poly
            this._pointA[0] = gO._adjusted_points[i][0]*s + gO._projected2dPosX;
            this._pointA[1] = gO._adjusted_points[i][1]*s + gO._projected2dPosY;

            if (i < gO._adjusted_points.length - 1) {   
                this._pointB[0] = gO._adjusted_points[i + 1][0]*s + gO._projected2dPosX;
                this._pointB[1] = gO._adjusted_points[i + 1][1]*s + gO._projected2dPosY;
            } else {
                this._pointB[0] = gO._adjusted_points[0][0]*s + gO._projected2dPosX;
                this._pointB[1] = gO._adjusted_points[0][1]*s + gO._projected2dPosY;
            }

            /*
            DEBUG_DISPLAY._debugOverlayContext.beginPath();
            DEBUG_DISPLAY._debugOverlayContext.strokeStyle = 'rgba(255,0,0,1.0)';
            DEBUG_DISPLAY._debugOverlayContext.lineWidth = 4;
            DEBUG_DISPLAY._debugOverlayContext.moveTo(this._pointA[0], this._pointA[1]);
            DEBUG_DISPLAY._debugOverlayContext.lineTo(this._pointB[0], this._pointB[1]);
            DEBUG_DISPLAY._debugOverlayContext.stroke();
            DEBUG_DISPLAY._debugOverlayContext.lineWidth = 1;
            */

            for (var j = 0; j < gO2._adjusted_points.length; j++) {
                
                //get line segment C - D from second poly
                this._pointC[0] = gO2._adjusted_points[j][0]*s + gO2._projected2dPosX;
                this._pointC[1] = gO2._adjusted_points[j][1]*s + gO2._projected2dPosY;

                if (j < gO2._adjusted_points.length - 1) {
                    this._pointD[0] = gO2._adjusted_points[j + 1][0]*s + gO2._projected2dPosX;
                    this._pointD[1] = gO2._adjusted_points[j + 1][1]*s + gO2._projected2dPosY;
                } else {
                    this._pointD[0] = gO2._adjusted_points[0][0]*s + gO2._projected2dPosX;
                    this._pointD[1] = gO2._adjusted_points[0][1]*s + gO2._projected2dPosY;
                }

                //check for line/line intersection, and break if 
                //we find one
                bIntersects = this.lineLineCollisionCheck();
                
                if (bIntersects) {
                    return true;
                }
            }
        }
        return false;
    }
    //--------------------------------------------------------------------------

    //-------[ lineLineCollisionCheck ]-----------------------------------------
    /* Purpose: checks for collision between two lines.  Used by the poly-based
    collision methods to check each segment of a poly against the segments of
    the other.
    */
    //--------------------------------------------------------------------------
    lineLineCollisionCheck() {
        var f = ((this._pointB[0] - this._pointA[0]) *
                 (this._pointD[1] - this._pointC[1]) -
                 (this._pointB[1] - this._pointA[1]) *
                 (this._pointD[0] - this._pointC[0]));

        if (f == 0) { return false; } //parallel or co-incident

        var d = ((this._pointA[1] - this._pointC[1]) *
                 (this._pointD[0] - this._pointC[0]) -
                 (this._pointA[0] - this._pointC[0]) *
                 (this._pointD[1] - this._pointC[1]));
        
        if (f > 0) { //alpha tests
            if (d < 0 || d > f) {
                return false;
            }
        } else {
            if (d > 0 || d < f) {
                return false;
            }
        }

        var e = ((this._pointA[1] - this._pointC[1]) *
                 (this._pointB[0] - this._pointA[0]) -
                 (this._pointA[0] - this._pointC[0]) *
                 (this._pointB[1] - this._pointA[1]));

        if (f > 0) { //beta tests
            if (e < 0 || e > f) {
                return false;
            }
        } else {
            if (e > 0 || e < f) {
                return false;
            }
        }

        return true;
    }
    //--------------------------------------------------------------------------


    //-------[ circlePolyCollisionCheck ]---------------------------------------
    /* Purpose: checks for collision between two game objects, one of which is
    type circle, and one of which is an arbitrary n-gon */
    //--------------------------------------------------------------------------
    circlePolyCollisionCheck(gOCircle, gOPoly) {
        var s = GAME_WINDOW.GAME_WINDOW_SCALE;
        var r = gOCircle._radius * s;

        //check to see if any points of the poly are inside the circle
        for (var i = 0; i < gOPoly._adjusted_points.length; i++) {
            this._pointA[0] = gOPoly._adjusted_points[i][0] * s + gOPoly._projected2dPosX;
            this._pointA[1] = gOPoly._adjusted_points[i][1] * s + gOPoly._projected2dPosY;

            var dX = this._pointA[0] - gOCircle._projected2dPosX;
            var dY = this._pointA[1] - gOCircle._projected2dPosY;
            dX = dX*dX;
            dY = dY*dY;
            
            if (dX + dY < r*r) {

                return true;
            }
        }

        //check to see if the center of the poly object is inside circle radius
        //a basic attempt at "containment"
        this._pointA[0] = gOPoly._projected2dPosX;
        this._pointA[1] = gOPoly._projected2dPosY;
        dX = this._pointA[0] - gOCircle._projected2dPosX;
        dY = this._pointA[1] - gOCircle._projected2dPosY;
        dX = dX * dX;
        dY = dY * dY;
        if (dX + dY < r * r) {

            return true;
        }
        return false;
    }
    //--------------------------------------------------------------------------


    //-------[ circleLineCollisionCheck2 ]---------------------------------------
    /* Purpose: checks for collision between a circle and a line.  Assumes that
    line AB has been packed into reusable object points */
    //--------------------------------------------------------------------------
    circleLineCollisionCheck2(radius) {
        var LAB = Math.sqrt((this._pointB[0] - this._pointA[0]) *
                            (this._pointB[0] - this._pointA[0]) +
                            (this._pointB[1] - this._pointA[1]) *
                            (this._pointB[1] - this._pointA[1]));

        var Dx = (this._pointB[0] - this._pointA[0]) / LAB;
        var Dy = (this._pointB[1] - this._pointA[1]) / LAB;

        var t = Dx * (this._pointC[0] - this._pointA[0]) +
                Dy * (this._pointC[1] - this._pointA[1]);

        var Ex = t * Dx + this._pointA[0];
        var Ey = t * Dy + this._pointA[1];

        var LEC = Math.sqrt((Ex - this._pointC[0]) *
            (Ex - this._pointC[0]) + (Ey - this._pointC[1]) *
            (Ey - this._pointC[1]));

        if (LEC < radius) {
            var dt = Math.sqrt(radius * radius - LEC * LEC);
            if (t - dt < 0 || t - dt > 1) {
                return true;
            }
        }
        return false;
    }
    //--------------------------------------------------------------------------


    //-------[ circleLineSegmentCollisionCheck ]--------------------------------
    /* Purpose: checks for collision between a circle and a line segment.  */
    //--------------------------------------------------------------------------
    circleLineSegmentCollisionCheck(cX, cY, cR, x1, y1, x2, y2) {

        var dX = x2 - x1;
        var dY = y2 - y1;
        var A = (dX * dX + dY * dY);
        var B = 2 * (dX * (x1 - cX) + dY * (y1 - cY));
        var C = (x1 - cX) * (x1 - cX) + (y1 - cY) * (y1 - cY) - cR * cR;
        var det = B * B - 4 * A * C;

        if ((A <= 0.0000001) || (det < 0)) {
            //no real solutions.
            return false;
        }

        //now we completely fake this based on the relative scale of
        //ships and lasers rather than truly intersect..., but this
        //is waaaaay cheaper
        if (Math.abs(x1 - cX) < cR && Math.abs(y1 - cY) < cR) {
            return true;
        }
        if (Math.abs(x2 - cX) < cR && Math.abs(y2 - cY) < cR) {
            return true;
        }
    }
    //--------------------------------------------------------------------------


    //-------[ circleLineCollisionCheck ]---------------------------------------
    /* Purpose: checks for collision between a circle and a line.  Assumes that
    line AB has been packed into reusable object points */
    //--------------------------------------------------------------------------
    circleLineCollisionCheck(radius) {
        // compute the triangle area times 2 (area = area2/2)
        var area2 = Math.abs((this._pointB[0] - this._pointA[0]) *
                             (this._pointC[1] - this._pointA[1]) -
                             (this._pointC[0] - this._pointA[0]) *
                             (this._pointB[1] - this._pointA[1]));

        // compute the AB segment length
        var LAB = Math.sqrt((this._pointB[0] - this._pointA[0]) *
                            (this._pointB[0] - this._pointA[0]) +
                            (this._pointB[1] - this._pointA[1]) *
                            (this._pointB[1] - this._pointA[1]));

        // compute the triangle height
        var h = area2 / LAB;

        // if the line intersects the circle
        if( h < radius ) {
            return true;
        } else {
            return false;
        }
    }
    //--------------------------------------------------------------------------


    //-------[ circleSquareCollisionCheck ]-------------------------------------
    /* Purpose: checks for collision between two game objects, one of which is
    type circule, and one of which is type square (in that order) */
    //--------------------------------------------------------------------------
    circleSquareCollisionCheck(gOCircle, gORect) {
        //console.log("CIRCLE SQUARE CHECK ");
        var cx = Math.abs(gOCircle._2dPosX - gORect._2dPosX);
        var xDist = gORect._hw + gOCircle._radius;
        if (cx > xDist) { return false; }

        var cy = Math.abs(gOCircle._2dPosY - gORect._2dPosY);
        var yDist = gORect._hh + gOCircle._radius;
        if (cy > yDist) { return false; }

        if (cx <= gORect._hw || cy <= gORect._hh) { return true; }

        var xCornerDist = cx - gORect._hw;
        var yCornerDist = cy - gORect._hh;
        var xCDS = xCornerDist * xCornerDist;
        var yCDS = yCornerDist * yCornerDist;
        var maxCDS = gOCircle._radius * gOCircle._radius;

        if (xCDS + yCDS <= maxCDS) { return true; }
        return false;
    }
}
//-------------------------------------------------------------------------------


//-------[ BIN ]-----------------------------------------------------------------
/* Purpose: a 'bin object' that is used to partition the screen into various
chunks and store game objects that happen to be inside those coordinates on
a given frame.  Game objects are in sublists by type (players, bullets, enemies, 
enemy bullets, etc.)  Futher documentation on type is in the gameObject class.
*/
//-------------------------------------------------------------------------------
function Bin(x1, y1, x2, y2, manager) {
    this._x1 = x1;
    this._y1 = y1;
    this._x2 = x2;
    this._y2 = y2;
    this._gOs = [[],[],[],[]];
    //this._manager = manager;
    return this;
}
//-------------------------------------------------------------------------------