﻿//-------[ GAME WINDOW ]---------------------------------------------------
/*  
    PURPOSE: Class for a GAME_WINDOW object that contains initialization, 
    resize, methods to keep the game window at a set aspect ratio regardless
    of the containing window ratio.
*/
//-------------------------------------------------------------------------
"use strict";

//-------------------------------------------------------------------------
var GAME_WINDOW_UNSCALED_RESOLUTION = 1920; //the base scale of unscaled coords,
var GAME_WINDOW_GLOBAL_SCALE = 1.0;
//this value is an arbitrary size that is scaled to fit the actual game 
//window.  The entire game world is viewed from this coordinate scale.
//NOTE: this var is the resolution of the game window orientation axis 
//(horiz / vert), with the resulting other axis size defined by aspect ratio.
//-------------------------------------------------------------------------


//-------------------------------------------------------------------------
var GAME_WINDOW_ZOOM_EFFECT = 1.2; //effect z-coord has on scale.  1 is default
//greater than one has an outsized effect, smaller than 1 minimizes effect
//0.0 sets everything to the z=0 plane regardless of z coord
//-------------------------------------------------------------------------


//-------------------------------------------------------------------------
class GameWindow {
    constructor() {

        //pattern to ensure singleton of GAME_WINDOW object
        //there can be only one...
        if (!GAME_WINDOW) {
            GAME_WINDOW = this;
        } else {
            console.log("AAENGINE: ERROR: Attempt to create multiple game windows." +
                      "\nAAENGINE: ERROR: Only first instanciation will be used.");
            return GAME_WINDOW;
        }
    };

    //-------[ GAME_WINDOW::INIT ]----------------------------------------
    /*
        PURPOSE: Called at startup or to re-initialize the game window
        to a specific aspect ratio and orientation
    */
    //--------------------------------------------------------------------
    init(aspectW, aspectH, orientation) {
        console.log("AAENGINE: Initializing Game Window...");
        this._engineBackground = document.getElementById("engine_background");
        this._PLAYFIELD_ORIENTATION = orientation; // 'vertical' or 'horizontal'
        this._VIEW_ASPECT_WIDTH = aspectW; //horizontal aspect (is relative to vertical)
        this._VIEW_ASPECT_HEIGHT = aspectH; //vertical aspect (is relative to horizontal)
        this._VIEW_ASPECT = this._VIEW_ASPECT_WIDTH / this._VIEW_ASPECT_HEIGHT;

        this.resize(); //call resize to set window relative to these values
    }

    //-------[ GAME_WINDOW::RESIZE ]--------------------------------------
    /*
        PURPOSE: Called in response to a resize event, resets exact location
        of engine_background div and elements contained therein to specific
        size based on ratio and window elements.
    */
    //--------------------------------------------------------------------
    resize(e) {

        this._lastWindowWidth = this._WINDOW_WIDTH;
        this._lastWindowHeight = this._WINDOW_HEIGHT;

        //get width and height as one or both have changed
        var w = window.innerWidth;
        var h = window.innerHeight;

        //initialize other temporary variables
        var ah = h;
        var aw = w;
        var l = 0;
        var t = 0;

        //if playfield orientation is wide
        if (this._PLAYFIELD_ORIENTATION == 'horizontal') {

            //set window to as wide as the screen
            this._engineBackground.style.left = 0 + 'px';
            this._engineBackground.style.width = w + 'px';
            this._GAME_WINDOW_WIDTH = w;

            //calculate aspect height based on 16:9 aspect ratio
            ah = w * ((this._VIEW_ASPECT_HEIGHT * 1.0) /
                       this._VIEW_ASPECT_WIDTH);
            if (ah < h) {
                //if aspect height is less than the height of window
                t = (Math.floor((h * 0.5) - ah * 0.5));
                this._engineBackground.style.top =  t + 'px';
                this._engineBackground.style.height = Math.floor(ah) + 'px';
                this._GAME_WINDOW_HEIGHT = ah;
                this._engineBackground.style.width = w;
                this._GAME_WINDOW_WIDTH = w;
            } else {
                //set height explicitly and find a new adjusted width
                this._engineBackground.style.top = t + 'px';
                this._engineBackground.style.height = h + 'px';
                this._GAME_WINDOW_HEIGHT = h;

                aw = h * ((this._VIEW_ASPECT_WIDTH * 1.0) /
                           this._VIEW_ASPECT_HEIGHT);
                l = (Math.floor(w * 0.5 - aw * 0.5));
                this._engineBackground.style.left = l + 'px';
                this._engineBackground.style.width = Math.floor(aw) + 'px';
                this._GAME_WINDOW_WIDTH = aw;
            }
        }
        else if (this._PLAYFIELD_ORIENTATION == 'vertical') {
            //set window to as tall as the screen
            this._engineBackground.style.top = 0 + 'px';
            this._engineBackground.style.height = h + 'px';
            this._GAME_WINDOW_HEIGHT = h;

            aw = h * ((this._VIEW_ASPECT_WIDTH * 1.0) /
                       this._VIEW_ASPECT_HEIGHT);
            if (aw < w) {
                //calculate aspect width based on 16:9 aspect ratio
                l = (Math.floor((w * 0.5) - aw * 0.5))
                this._engineBackground.style.left = l + 'px';
                this._engineBackground.style.width = Math.floor(aw) + 'px';
                this._GAME_WINDOW_WIDTH = aw;

            } else {
                //set width explicitly and find a new adjusted height
                this._engineBackground.style.width = w + 'px';
                this._GAME_WINDOW_WIDTH = w;
                this._engineBackground.style.left = l + 'px';
                ah = w * ((this._VIEW_ASPECT_HEIGHT * 1.0) /
                           this._VIEW_ASPECT_WIDTH);
                t = (Math.floor(h * 0.5 - ah * 0.5));
                this._engineBackground.style.top =  t + 'px';
                this._engineBackground.style.height = ah + 'px';
                this._GAME_WINDOW_HEIGHT = ah;
            }
        } else {
            console.log("AAENGINE: ERROR: Resize on window without set orientation.")
        }

        //set game window variables for reporting functions...
        this._WINDOW_WIDTH = w;
        this._WINDOW_HEIGHT = h;
        this._GAME_WINDOW_TOP = t;
        this._GAME_WINDOW_LEFT = l;

        if (this._PLAYFIELD_ORIENTATION == 'horizontal') {
            this._GAME_WINDOW_SCALE = this._GAME_WINDOW_WIDTH /
                  GAME_WINDOW_UNSCALED_RESOLUTION *
                  GAME_WINDOW_GLOBAL_SCALE;
            this._UNSCALED_WIDTH = GAME_WINDOW_UNSCALED_RESOLUTION /
                GAME_WINDOW_GLOBAL_SCALE;
            this._UNSCALED_HEIGHT = GAME_WINDOW_UNSCALED_RESOLUTION /
                GAME_WINDOW_GLOBAL_SCALE / this._VIEW_ASPECT;
        } else {
            this._GAME_WINDOW_SCALE = this._GAME_WINDOW_HEIGHT /
                  GAME_WINDOW_UNSCALED_RESOLUTION * 
                  GAME_WINDOW_GLOBAL_SCALE;
            this._UNSCALED_WIDTH = GAME_WINDOW_UNSCALED_RESOLUTION /
                GAME_WINDOW_GLOBAL_SCALE * this._VIEW_ASPECT;
            this._UNSCALED_HEIGHT = GAME_WINDOW_UNSCALED_RESOLUTION /
                GAME_WINDOW_GLOBAL_SCALE;
        }

        this._GAME_WINDOW_HALF_WIDTH = this._GAME_WINDOW_WIDTH*0.5;
        this._GAME_WINDOW_HALF_HEIGHT = this._GAME_WINDOW_HEIGHT*0.5;
        this._GAME_WINDOW_UNSCALED_HALF_WIDTH = this._UNSCALED_WIDTH*0.5;
        this._GAME_WINDOW_UNSCALED_HALF_HEGIHT = this._UNSCALED_HEIGHT*0.5;

        //make input adjustments
        INPUT_MANAGER.resize(e);

        //resize the 3D canvas
        AA3D_MANAGER.resize();

        //if a debug display exists, resize debug display
        if (DEBUG_DISPLAY) {
            DEBUG_DISPLAY.resize();
        }

        //update thread copies of this information if we are threading
        if (USE_THREADING) {
            THREAD_MANAGER.resize();
        }

        //if we are debuging, set the slice scale based on window size
        if (DEBUG) {
            if (FPS_COUNTER) { FPS_COUNTER.setSliceScale(); }
        }
    }

    //--------------------------------------------------------------------
    //accessor methods for game window properties that are used elsewhere
    //NOTE: private here is by naming convention only
    //--------------------------------------------------------------------
    get WINDOW_WIDTH() { return this._WINDOW_WIDTH; }
    get WINDOW_HEIGHT() { return this._WINDOW_HEIGHT; }
    get GAME_WINDOW_WIDTH() { return this._GAME_WINDOW_WIDTH; }
    get GAME_WINDOW_HEIGHT() { return this._GAME_WINDOW_HEIGHT; }
    get GAME_WINDOW_LEFT() { return this._GAME_WINDOW_LEFT; }
    get GAME_WINDOW_TOP() { return this._GAME_WINDOW_TOP; }
    get GAME_WINDOW_SCALE() { return this._GAME_WINDOW_SCALE; }
    get VIEW_ASPECT() { return this._VIEW_ASPECT; }
    get GAME_WINDOW_UNSCALED_WIDTH() { return this._UNSCALED_WIDTH; }
    get GAME_WINDOW_UNSCALED_HEIGHT() { return this._UNSCALED_HEIGHT; }
    get GAME_WINDOW_HALF_WIDTH() { return this._GAME_WINDOW_HALF_WIDTH; }
    get GAME_WINDOW_HALF_HEIGHT() { return this._GAME_WINDOW_HALF_HEIGHT; }
    get GAME_WINDOW_UNSCALED_HALF_WIDTH() { return this._GAME_WINDOW_UNSCALED_HALF_WIDTH; }
    get GAME_WINDOW_UNSCALED_HALF_HEIGHT() { return this._GAME_WINDOW_UNSCALED_HALF_HEGIHT; }

}
//-------------------------------------------------------------------------


//-------------------------------------------------------------------------
// creation of global GAME_WINDOW reference so it cam be bound
// to the window.onresize event.  Must occur after class definition
//-------------------------------------------------------------------------
var GAME_WINDOW = GAME_WINDOW || new GameWindow();
//-------------------------------------------------------------------------


//-------[ RESIZE ]--------------------------------------------------------
/*  
    PURPOSE: called whenever main game window is resized (either by user
    in a browser context or by changing screen resolution of the monitor
    in full screen scenarios)  NOTE: called on init as well which sets
    aspect ratios and orientation.
*/
//-------------------------------------------------------------------------
window.addEventListener('resize', function (e) {
    if (GAME_WINDOW) {
        GAME_WINDOW.resize(e);
    }
}, false);
//-------------------------------------------------------------------------




