﻿//-------[ DEBUG DISPLAY ]-------------------------------------------------
/*  
    PURPOSE: Class for a heads up debug display.  Binds to a series of 
    elements in the main HTML file in the debug_display div.  Makes
    use of the debug_display.css class and style definitions.

    //-----[ TUNING VARIABLES ]--------------------------------------------
    1) DISPLAY_UPDATE_SLICING - this tells the debug display to update all
    tracked fields every frame (when false) or to 'round robin' its updates
    such that mouse input is frame 1, keyboard input is frame 2, and controller
    input is frame 3.  This is simply a performance optimization.

    2) DISPLAY_UPDATE_FRAME_EQUALIZATION - this tells the debug display
    whether to run as fast as possible every frame, or to 'pad' the time 
    taken to a threshold value set in this._timingThreshold in init(). 
    (Currently set at 1.0ms).  This is useful as running as fast as possible
    will produce 'thrashing' in terms of time as asynchonous input events 
    may or may not fire every frame, and thus some frames will have updates
    while others won't.  For time slicing other elements of the application
    while also wrapping the update of this display in a slice, it is sometimes
    useful to hold the timing of this activity to a relative constant.  

    //-----[ AESTHETIC VARIABLES ]------------------------------------------
    1) this._backgroundHighlightClr is defined in init().  This is the color
    used for the background of fields when their values change.

    2) this._backgroundClr is defined in init().  This is the color used for
    the background of value fields.

*/
//-------------------------------------------------------------------------
"use strict";
//-------------------------------------------------------------------------
var DISPLAY_UPDATE_SLICING = true;  //Y/N update select items per frame
var DISPLAY_UPDATE_FRAME_EQUALIZATION = false; //Y/N pad to time threshold
var DEBUG_CONSOLE = null; //console to print debug information
var ECHO_NORMAL_CONSOLE = true; //boolean Y/N echo log to normal console
//-------------------------------------------------------------------------
// 
//-------------------------------------------------------------------------


//-------------------------------------------------------------------------
class DebugDisplay {
    //------[  DEBUG DISPLAY::Constructor ]--------------------------------
    // Purpose: Create the debug display object and base values
    //---------------------------------------------------------------------
    constructor() {
        
        DEBUG_CONSOLE = console.log; //overwrite of console functionality
        this._debugConsoleLineHeight = 19; //number of text lines in console
        this._debugConsoleMsgs = [];

        this._backgroundHighlightClr = 'rgba(50,200,255,0.85)';
        this._backgroundClr = 'rgba(100,100,100, 0.5)';

        this._mX = 0; //temporary storage of last value for display check
        this._mY = 0; //temporary storage of last value for display check
        this._wD = 0; //temporary storage of last value for display check
        this._keyboardEventFrame = true; //is there a kb event this frame
        this._mouseEventFrame = true; //is there a mouse event this frame
        this._mouseMoveEventFrameX = true; //mouseX move event this frame
        this._mouseMoveEventFrameY = true; //mouseY move event this frame
        this._mouseWheelEventFrame = true; //mouse wheel event this frame

        //gamepad data stores for first gamepad
        this._lastGP1con = -1;
        this._lastGP1map = 0;
        this._updateGP1 = []
        for (var i = 0; i < 23; i++) {
            this._updateGP1[i] = true;
        }

        //gamepad data stores for second gamepad
        this._lastGP2con = -1;
        this._lastGP2map = 0;
        this._updateGP2 = []
        for (var i = 0; i < 23; i++) {
            this._updateGP2[i] = true;
        }

        //pattern to ensure singleton of GAME_WINDOW object
        //there can be only one...
        if (!DEBUG_DISPLAY) {
            DEBUG_DISPLAY = this;
            
        } else {
            console.log(
            "AAENGINE: ERROR: Attempt to create multiple debug displays." +
            "\nAAENGINE: ERROR: Only first instanciation will be used.");
            return DEBUG_DISPLAY;
        }
    };
    //---------------------------------------------------------------------

    //------[  DEBUG DISPLAY::Init]----------------------------------------
    // Purpose: set initial values for counters and logic states
    //---------------------------------------------------------------------
    init() {
        this._timingThreshold = 1.0; //time padding to avoid frame thrashing
        this._updateTime = 0;        //time taken to update the frame
        this._lastMouseXValue = 0;   //the last mouseX value 
        this._lastMouseYValue = 0;   //the last mouseY value 
        this._lastMouseWheelDelta = 0; //the last value for mouse wheel delta
        this._renderCounter = 0; //the counter of which render pass to use
        this.setDebugHandles(); //call to setDebugHandles to grab elements 
        this._debugBins = []; //if threaded, store bins internal to this

        /*
        for (var i = 0; i < COLLISION_BINS_X; i++) {
            this._debugBins[i] = [];
            for (var j = 0; j < COLLISION_BINS_Y;) {
                this._debugBins[i][j] = 0;
            }
        }*/
        this._2dDebugVisibility = true;

        this._DebugConsoleWin = document.getElementById('debug_console');
        
        this._debugText = document.getElementById("debugConsoleText");
        var a = [];
        for (var i = 0; i < this._debugConsoleLineHeight; i++) {
            var e = document.createTextNode(""); this._debugText.appendChild(e);
            a[i] = e;
            e = document.createElement("br"); this._debugText.appendChild(e);
        }
        this._debugText = a;

        if (USE_THREADING) {
            this._threadsEnabledDisplay.nodeValue = "true";
            this._threadsEnabledDisplayBackground.background = this._backgroundHighlightClr;
        }
    }
    //---------------------------------------------------------------------


    //------[  DEBUG DISPLAY::resize]--------------------------------------
    // Purpose: actions to keep alignment of various windows and displays
    //when the window is resized, namely the console and the debug overlay  
    //---------------------------------------------------------------------
    resize() {
        if (this._visibility) {
            this._DebugConsoleWin.style.top =
                GAME_WINDOW.WINDOW_HEIGHT - 326 + 'px';
            this._DebugConsoleWin.style.left =
                GAME_WINDOW.WINDOW_WIDTH - 730 + 'px';
            this._DebugConsoleWin.style.width = 700 + 'px';
            this._DebugConsoleWin.style.visibility = 'visible';

            if (GAME_WINDOW.WINDOW_WIDTH < 720 + 700 + 66) {
                var l = parseInt(window.getComputedStyle(
                    this._debugDisplay).width) + 36;
                this._DebugConsoleWin.style.left = l + 'px';
                var w = GAME_WINDOW.WINDOW_WIDTH - l - 30;
                if (w < 300) {
                    this._DebugConsoleWin.style.visibility = 'hidden';
                } else {
                    this._DebugConsoleWin.style.width = w + 'px';
                    this._DebugConsoleWin.style.visibility = 'visible';
                }
            }
        }

        if (this._debugOverlayCanvas) {
            this._debugOverlayCanvas.style.width =
                GAME_WINDOW.GAME_WINDOW_WIDTH + 'px';
            this._debugOverlayCanvas.style.height =
                GAME_WINDOW.GAME_WINDOW_HEIGHT + 'px';

            this._debugOverlayCanvas.width =
                GAME_WINDOW.GAME_WINDOW_WIDTH;
            this._debugOverlayCanvas.height =
                GAME_WINDOW.GAME_WINDOW_HEIGHT;
        }
    }
    //---------------------------------------------------------------------


    //------[  DEBUG DISPLAY::initDebugOverlay]----------------------------
    // Purpose: creates a debug overlay for the entire scene, that can be
    // drawn to to visualize 2D objects as needed (collision boxes, vectors,
    // etc.
    //---------------------------------------------------------------------
    initDebugOverlay() {
        var e = document.getElementById('engine_background');
        this._debugOverlayCanvas = document.createElement('canvas');
        this._debugOverlayContext = this._debugOverlayCanvas.getContext('2d');
        this._debugOverlayCanvas.width = GAME_WINDOW.GAME_WINDOW_WIDTH;
        this._debugOverlayCanvas.height = GAME_WINDOW.GAME_WINDOW_HEIGHT;
        e.appendChild(this._debugOverlayCanvas);

        this._debugOverlayCanvas.style.width = GAME_WINDOW.GAME_WINDOW_WIDTH + 'px';
        this._debugOverlayCanvas.style.height = GAME_WINDOW.GAME_WINDOW_HEIGHT + 'px';
        this._debugOverlayCanvas.style.top = 0 + 'px';
        this._debugOverlayCanvas.style.left = 0 + 'px';
        this._debugOverlayCanvas.style.position = 'relative';
        this._debugOverlayCanvas.style.background = 'rgba(0,128,0,0.0)'
    }
    //---------------------------------------------------------------------


    //------[  DEBUG DISPLAY::toggleDebugDisplay]--------------------------
    // Purpose: set the visibility of the debug display and turn rendering
    // on and off via the visibility flag.  
    //---------------------------------------------------------------------
    toggleDebugDisplay() {
        if (this._visibility) {
            this._debugDisplay.style.visibility = 'hidden';
            this._DebugConsoleWin.style.visibility = 'hidden'
            this._visibility = false;
        } else {
            this._debugDisplay.style.visibility = 'visible';
            this._DebugConsoleWin.style.visibility = 'visible'
            this._visibility = true;
        }

        this.resize();
    }
    //---------------------------------------------------------------------
    

    //-----[ DEBUG DISPLAY::toggleDebugCanvas]-----------------------------
    toggleDebugCanvas() {
        if (DEBUG) {
            if (this._2dDebugVisibility) {
                this._debugOverlayCanvas.style.visibility = 'hidden';
                this._2dDebugVisibility = false;
                if (USE_THREADING) {
                    THREAD_MANAGER.updateDebugCanvasState(false);
                }
            } else {
                this._debugOverlayCanvas.style.visibility = 'visible';
                this._2dDebugVisibility = true;
                if (USE_THREADING) {
                    THREAD_MANAGER.updateDebugCanvasState(true);
                }
            }
        }
    }
    //---------------------------------------------------------------------


    //------[  DEBUG DISPLAY::Update]--------------------------------------
    // Purpose: main workhorse of the debug display, called per frame from
    // the main applicaiton loop.  
    //---------------------------------------------------------------------
    update() {

        var t = performance.now(); // set the current time at the start 

        if (this._visibility == true) {

            //set the background color of the display for the use_threads 
            //tile.  do this separate as it marks the whole thing as "dirty"
            //so it will be painted.  Win/WinJS has some issues on whether
            //it updates the rendered page if only nodeValues change
            if (USE_THREADING) {
                this._threadsEnabledDisplayBackground = this._backgroundHighlightClr;
            } else {
                this._threadsEnabledDisplayBackground = this._backgroundClr;
            }

            //since this is expensive in some browsers, 
            //update this every few frames instead of every frame
            if (this._renderCounter == 0 || !DISPLAY_UPDATE_SLICING) {
                //---------------------------------------------------------
                //update mouse X position and background
                //---------------------------------------------------------
                this._mX = INPUT_MANAGER.MOUSE_X;
                if (this._mX != this._lastMouseXValue) {
                        this._mouseXvalue.nodeValue = this._mX;
                        this._lastMouseXValue = this._mX;
                        this._mouseXlastValue.nodeValue =
                            INPUT_MANAGER.LAST_MOUSE_X;
                        this._mouseXrawValue.nodeValue =
                            INPUT_MANAGER.RAW_MOUSE_X;
                        this._mouseXBackground.background =
                            this._backgroundHighlightClr;
                        this._mouseXlastValueBackground.background =
                            this._backgroundHighlightClr;
                        this._mouseXrawValueBackground.background =
                            this._backgroundHighlightClr;
                        this._mouseMoveEventFrameX = true;

                        if (INPUT_MANAGER.MOUSE_ACTIVE > 0) {
                            this._mouseActiveValue.nodeValue = "1";
                            this._mouseActiveValueBackground.background =
                                this._backgroundHighlightClr;
                        } else if (INPUT_MANAGER.MOUSE_ACTIVE == 0) {
                            this._mouseActiveValue.nodeValue = "0";
                            this._mouseActiveValueBackground.background =
                                this._backgroundClr;
                        } else {
                            this._mouseActiveValue.nodeValue = "-1";
                            this._mouseActiveValueBackground.background =
                                this._backgroundClr;
                        }

               } else {
                   if (this._mouseMoveEventFrameX) {
                       this._mouseXBackground.background =
                           this._backgroundClr;
                       this._mouseXlastValueBackground.background =
                           this._backgroundClr;
                       this._mouseXrawValueBackground.background =
                           this._backgroundClr;
                       this._mouseMoveEventFrameX = false;
                   }
                }
                //---------------------------------------------------------
                //update mouse Y position and background
                //---------------------------------------------------------
                this._mY = INPUT_MANAGER.MOUSE_Y;
                if (this._mY != this._lastMouseYValue) {
                        this._mouseYvalue.nodeValue = this._mY;
                        this._lastMouseYValue = this._mY;
                        this._mouseYlastValue.nodeValue =
                            INPUT_MANAGER.LAST_MOUSE_Y;
                        this._mouseYrawValue.nodeValue =
                            INPUT_MANAGER.RAW_MOUSE_Y;
                        this._mouseYBackground.background =
                            this._backgroundHighlightClr;
                        this._mouseYlastValueBackground.background =
                            this._backgroundHighlightClr;
                        this._mouseYrawValueBackground.background =
                            this._backgroundHighlightClr;
                        this._mouseMoveEventFrameY = true;

                        if (INPUT_MANAGER.MOUSE_ACTIVE > 0) {
                            this._mouseActiveValue.nodeValue = "1";
                            this._mouseActiveValueBackground.background =
                                this._backgroundHighlightClr;
                        } else if (INPUT_MANAGER.MOUSE_ACTIVE == 0) {
                            this._mouseActiveValue.nodeValue = "0";
                            this._mouseActiveValueBackground.background =
                                this._backgroundClr;
                        } else {
                            this._mouseActiveValue.nodeValue = "-1";
                            this._mouseActiveValueBackground.background =
                                this._backgroundClr;
                        }

                } else {
                    if (this._mouseMoveEventFrameY) {
                        this._mouseYBackground.background =
                            this._backgroundClr;
                        this._mouseYlastValueBackground.background =
                            this._backgroundClr;
                        this._mouseYrawValueBackground.background =
                            this._backgroundClr;
                        this._mouseMoveEventFrameY = false;
                    }
                }

                //---------------------------------------------------------
                // update mouse move events
                //---------------------------------------------------------
                if (this._mouseEventFrame) {
      
                    if (INPUT_MANAGER.MOUSE_LEFT) {
                        this._mouseLeftValue.nodeValue = "1";
                        this._mouseLeftValueBackground.background =
                            this._backgroundHighlightClr;
                    } else {
                        this._mouseLeftValue.nodeValue = "0";
                        this._mouseLeftValueBackground.background =
                            this._backgroundClr;
                    }
                    if (INPUT_MANAGER.MOUSE_MIDDLE) {
                        this._mouseMiddleValue.nodeValue = "1";
                        this._mouseMiddleValueBackground.background =
                            this._backgroundHighlightClr;
                    } else {
                        this._mouseMiddleValue.nodeValue = "0";
                        this._mouseMiddleValueBackground.background =
                            this._backgroundClr;
                    }
                    if (INPUT_MANAGER.MOUSE_RIGHT) {
                        this._mouseRightValue.nodeValue = "1";
                        this._mouseRightValueBackground.background =
                            this._backgroundHighlightClr;
                    } else {
                        this._mouseRightValue.nodeValue = "0";
                        this._mouseRightValueBackground.background =
                            this._backgroundClr;
                    }
                    
                }
                this._mouseEventFrame = false;

                //---------------------------------------------------------
                // update mouse Wheel events
                //---------------------------------------------------------
                if (this._mouseWheelEventFrame) {
                    //update other mosuse stuff
                    this._wD = INPUT_MANAGER.MOUSE_WHEEL_TOTAL_DELTA;
                    if (this._wD != this._lastMouseWheelDelta) {
                        this._mouseWheelDeltaValue.nodeValue = this._wD;
                        this._lastMouseWheelDelta = this._wD;
                        this._mouseWheelDeltaValueBackground.background =
                            this._backgroundHighlightClr;
                    } else {
                        this._mouseWheelDeltaValue.nodeValue = this._wD;
                        this._mouseWheelDeltaValueBackground.background =
                            this._backgroundClr;
                        this._mouseWheelEventFrame = false;
                    }
                    if (INPUT_MANAGER.MOUSE_WHEEL) {
                        this._mouseWheelEvent.nodeValue = "1";
                        this._mouseWheelEventBackground.background =
                            this._backgroundHighlightClr;
                    } else {
                        this._mouseWheelEvent.nodeValue = "0";
                        this._mouseWheelEventBackground.background =
                            this._backgroundClr;
                    }
                }
            }

            //-------------------------------------------------------------
            // secondary render path - keyboard updates
            //-------------------------------------------------------------
            if (this._renderCounter == 1 || !DISPLAY_UPDATE_SLICING) {
                
                if (this._keyboardEventFrame) {

                    if (INPUT_MANAGER.KB_SPACE) {
                        this._spaceValue.nodeValue = "1";
                        this._spaceValueBackground.background =
                            this._backgroundHighlightClr;
                    } else {
                        this._spaceValue.nodeValue = "0";
                        this._spaceValueBackground.background =
                            this._backgroundClr;
                    }

                    if (INPUT_MANAGER.KB_UP) {
                        this._KBUp.nodeValue = "1";
                        this._KBUpBackground.background =
                            this._backgroundHighlightClr;
                    } else {
                        this._KBUp.nodeValue = "0";
                        this._KBUpBackground.background =
                            this._backgroundClr;
                    }

                    if (INPUT_MANAGER.KB_DOWN) {
                        this._KBDown.nodeValue = "1";
                        this._KBDownBackground.background =
                            this._backgroundHighlightClr;
                    } else {
                        this._KBDown.nodeValue = "0";
                        this._KBDownBackground.background =
                            this._backgroundClr;
                    }

                    if (INPUT_MANAGER.KB_LEFT) {
                        this._KBLeft.nodeValue = "1";
                        this._KBLeftBackground.background =
                            this._backgroundHighlightClr;
                    } else {
                        this._KBLeft.nodeValue = "0";
                        this._KBLeftBackground.background =
                            this._backgroundClr;
                    }

                    if (INPUT_MANAGER.KB_RIGHT) {
                        this._KBRight.nodeValue = "1";
                        this._KBRightBackground.background =
                            this._backgroundHighlightClr;
                    } else {
                        this._KBRight.nodeValue = "0";
                        this._KBRightBackground.background =
                            this._backgroundClr;
                    }

                    this._keyboardEventFrame = false;
                }
                //render tracing updates
                this._callsDisplay.nodeValue =
                        AA3D_MANAGER.RENDERER.info.render.calls;
                this._verticesDisplay.nodeValue =
                        AA3D_MANAGER.RENDERER.info.render.vertices;
                this._facesDisplay.nodeValue =
                        AA3D_MANAGER.RENDERER.info.render.faces;
                this._pointsDisplay.nodeValue =
                        AA3D_MANAGER.RENDERER.info.render.points;
                this._numTexturesDisplay.nodeValue =
                        AA3D_MANAGER.RENDERER.info.memory.textures;
                this._numShadersDisplay.nodeValue =
                        AA3D_MANAGER.RENDERER.info.programs.length;
                
            }

            //-------------------------------------------------------------
            // tertiary render path - controller updates
            //-------------------------------------------------------------
            if (this._renderCounter == 2 || !DISPLAY_UPDATE_SLICING) {

                var controllerInput = INPUT_MANAGER.GAMEPAD1;

                //connection status
                if (controllerInput[0] != this._lastGP1con) {
                    this._GP1con.nodeValue = controllerInput[0];
                    this._lastGP1con = controllerInput[0];
                    if (controllerInput[0]) {
                        this._GP1conBackground.background =
                            this._backgroundHighlightClr;
                    } else {
                        this._GP1conBackground.background =
                            this._backgroundClr;
                    }
                }

                if (controllerInput[1] != this._lastGP1map) {
                    this._GP1map.nodeValue = controllerInput[1];
                    this._lastGP1map = controllerInput[1];
                    if (controllerInput[1]) {
                        this._GP1mapBackground.background =
                            this._backgroundHighlightClr;
                    } else {
                        this._GP1mapBackground.background =
                            this._backgroundClr;
                    }
                }

                //axes
                if (controllerInput[0]) {
                    //axes 0
                    if (controllerInput[3]) {
                        this._GP1leftHZ.nodeValue =
                            controllerInput[3].toPrecision(4);
                        this._GP1leftHZBackground.background =
                            this._backgroundHighlightClr;
                        this._updateGP1[3] = true;
                    } else if (this._updateGP1[3]) {
                        this._GP1leftHZBackground.background =
                            this._backgroundClr;
                        this._GP1leftHZ.nodeValue = "0.000";
                        this._updateGP1[3] = false;
                    }

                    //axes 1
                    if (controllerInput[4]) {
                        this._GP1leftVT.nodeValue =
                            controllerInput[4].toPrecision(4);
                        this._GP1leftVTBackground.background =
                            this._backgroundHighlightClr;
                        this._updateGP1[4] = true;
                    } else if (this._updateGP1[4]) {
                        this._GP1leftVTBackground.background =
                            this._backgroundClr;
                        this._GP1leftVT.nodeValue = "0.000";
                        this._updateGP1[4] = false;
                    }

                    //axes 2
                    if (controllerInput[5]) {
                        this._GP1rightHZ.nodeValue =
                            controllerInput[5].toPrecision(4);
                        this._GP1rightHZBackground.background =
                            this._backgroundHighlightClr;
                        this._updateGP1[5] = true;
                    } else if (this._updateGP1[5]) {
                        this._GP1rightHZBackground.background =
                            this._backgroundClr;
                        this._GP1rightHZ.nodeValue = "0.000";
                        this._updateGP1[5] = false;
                    }

                    //axes 3
                    if (controllerInput[6]) {
                        this._GP1rightVT.nodeValue =
                            controllerInput[6].toPrecision(4);
                        this._GP1rightVTBackground.background =
                            this._backgroundHighlightClr;
                        this._updateGP1[6] = true;
                    } else if (this._updateGP1[6]) {
                        this._GP1rightVTBackground.background =
                            this._backgroundClr;
                        this._GP1rightVT.nodeValue = "0.000";
                        this._updateGP1[6] = false;
                    }

                    //left trigger
                    if (controllerInput[13]) {
                        this._GP1LeftTrigger.nodeValue =
                            controllerInput[13].toPrecision(4);
                        this._GP1LeftTriggerBackground.background =
                            this._backgroundHighlightClr;
                        this._updateGP1[13] = true;
                    } else if (this._updateGP1[13]) {
                        this._GP1LeftTriggerBackground.background =
                            this._backgroundClr;
                        this._GP1LeftTrigger.nodeValue = "0.000";
                        this._updateGP1[13] = false;
                    }

                    //left trigger
                    if (controllerInput[14]) {
                        this._GP1RightTrigger.nodeValue =
                            controllerInput[14].toPrecision(4);
                        this._GP1RightTriggerBackground.background =
                            this._backgroundHighlightClr;
                        this._updateGP1[14] = true;
                    } else if (this._updateGP1[14]) {
                        this._GP1RightTriggerBackground.background =
                            this._backgroundClr;
                        this._GP1RightTrigger.nodeValue = "0.000";
                        this._updateGP1[14] = false;
                    }

                    //left bumper
                    if (controllerInput[11]) {
                        this._GP1LeftBumper.nodeValue =
                            controllerInput[11];
                        this._GP1LeftBumperBackground.background =
                            this._backgroundHighlightClr;
                        this._updateGP1[11] = true;
                    } else if (this._updateGP1[11]) {
                        this._GP1LeftBumperBackground.background =
                            this._backgroundClr;
                        this._GP1LeftBumper.nodeValue = "0";
                        this._updateGP1[11] = false;
                    }

                    //right bumper
                    if (controllerInput[12]) {
                        this._GP1RightBumper.nodeValue =
                            controllerInput[12];
                        this._GP1RightBumperBackground.background =
                            this._backgroundHighlightClr;
                        this._updateGP1[12] = true;
                    } else if (this._updateGP1[12]) {
                        this._GP1RightBumperBackground.background =
                            this._backgroundClr;
                        this._GP1RightBumper.nodeValue = "0";
                        this._updateGP1[12] = false;
                    }

                    //A BUTTON
                    if (controllerInput[7]) {
                        this._GP1a.nodeValue =
                            controllerInput[7];
                        this._GP1aBackground.background =
                            this._backgroundHighlightClr;
                        this._updateGP1[7] = true;
                    } else if (this._updateGP1[7]) {
                        this._GP1aBackground.background =
                            this._backgroundClr;
                        this._GP1a.nodeValue = "0";
                        this._updateGP1[7] = false;
                    }

                    //B BUTTON
                    if (controllerInput[8]) {
                        this._GP1b.nodeValue =
                            controllerInput[8];
                        this._GP1bBackground.background =
                            this._backgroundHighlightClr;
                        this._updateGP1[8] = true;
                    } else if (this._updateGP1[8]) {
                        this._GP1bBackground.background =
                            this._backgroundClr;
                        this._GP1b.nodeValue = "0";
                        this._updateGP1[8] = false;
                    }

                    //X BUTTON
                    if (controllerInput[9]) {
                        this._GP1x.nodeValue =
                            controllerInput[9];
                        this._GP1xBackground.background =
                            this._backgroundHighlightClr;
                        this._updateGP1[9] = true;
                    } else if (this._updateGP1[9]) {
                        this._GP1xBackground.background =
                            this._backgroundClr;
                        this._GP1x.nodeValue = "0";
                        this._updateGP1[9] = false;
                    }

                    //Y BUTTON
                    if (controllerInput[10]) {
                        this._GP1y.nodeValue =
                            controllerInput[10];
                        this._GP1yBackground.background =
                            this._backgroundHighlightClr;
                        this._updateGP1[10] = true;
                    } else if (this._updateGP1[10]) {
                        this._GP1yBackground.background =
                            this._backgroundClr;
                        this._GP1y.nodeValue = "0";
                        this._updateGP1[10] = false;
                    }

                    //DPAD U/D/L/R
                    if (controllerInput[19]) {
                        this._GP1dpUp.nodeValue =
                            controllerInput[19];
                        this._GP1dpUpBackground.background =
                            this._backgroundHighlightClr;
                        this._updateGP1[19] = true;
                    } else if (this._updateGP1[19]) {
                        this._GP1dpUpBackground.background =
                            this._backgroundClr;
                        this._GP1dpUp.nodeValue = "0";
                        this._updateGP1[19] = false;
                    }

                    if (controllerInput[20]) {
                        this._GP1dpDown.nodeValue =
                            controllerInput[20];
                        this._GP1dpDownBackground.background =
                            this._backgroundHighlightClr;
                        this._updateGP1[20] = true;
                    } else if (this._updateGP1[20]) {
                        this._GP1dpDownBackground.background =
                            this._backgroundClr;
                        this._GP1dpDown.nodeValue = "0";
                        this._updateGP1[20] = false;
                    }

                    if (controllerInput[21]) {
                        this._GP1dpLeft.nodeValue =
                            controllerInput[21];
                        this._GP1dpLeftBackground.background =
                            this._backgroundHighlightClr;
                        this._updateGP1[21] = true;
                    } else if (this._updateGP1[21]) {
                        this._GP1dpLeftBackground.background =
                            this._backgroundClr;
                        this._GP1dpLeft.nodeValue = "0";
                        this._updateGP1[21] = false;
                    }

                    if (controllerInput[22]) {
                        this._GP1dpRight.nodeValue =
                            controllerInput[22];
                        this._GP1dpRightBackground.background =
                            this._backgroundHighlightClr;
                        this._updateGP1[22] = true;
                    } else if (this._updateGP1[22]) {
                        this._GP1dpRightBackground.background =
                            this._backgroundClr;
                        this._GP1dpRight.nodeValue = "0";
                        this._updateGP1[22] = false;
                    }

                    //other misc buttons
                    if (controllerInput[16]) {
                        this._GP1start.nodeValue =
                            controllerInput[16];
                        this._GP1startBackground.background =
                            this._backgroundHighlightClr;
                        this._updateGP1[16] = true;
                    } else if (this._updateGP1[16]) {
                        this._GP1startBackground.background =
                            this._backgroundClr;
                        this._GP1start.nodeValue = "0";
                        this._updateGP1[16] = true;
                    }

                    if (controllerInput[15]) {
                        this._GP1back.nodeValue =
                            controllerInput[15];
                        this._GP1backBackground.background =
                            this._backgroundHighlightClr;
                        this._updateGP1[15] = true;
                    } else if (this._updateGP1[15]) {
                        this._GP1backBackground.background =
                            this._backgroundClr;
                        this._GP1back.nodeValue = "0";
                        this._updateGP1[15] = false;
                    }

                    if (controllerInput[17]) {
                        this._GP1leftStick.nodeValue =
                            controllerInput[17];
                        this._GP1leftStickBackground.background =
                            this._backgroundHighlightClr;
                        this._updateGP1[17] = true;
                    } else if (this._updateGP1[17]) {
                        this._GP1leftStickBackground.background =
                            this._backgroundClr;
                        this._GP1leftStick.nodeValue = "0";
                        this._updateGP1[17] = false;
                    }

                    if (controllerInput[18]) {
                        this._GP1rightStick.nodeValue =
                            controllerInput[18];
                        this._GP1rightStickBackground.background =
                            this._backgroundHighlightClr;
                        this._updateGP1[18] = true;
                    } else if (this._updateGP1[18]) {
                        this._GP1rightStickBackground.background =
                            this._backgroundClr;
                        this._GP1rightStick.nodeValue = "0";
                        this._updateGP1[18] = false;
                    }
                }
            }

            //-------------------------------------------------------------
            // quad render path - controller 2 updates
            //-------------------------------------------------------------
            if (this._renderCounter == 3 || !DISPLAY_UPDATE_SLICING) {

                var controllerInput = INPUT_MANAGER.GAMEPAD2;

                //connection status
                if (controllerInput[0] != this._lastGP2con) {
                    this._GP2con.nodeValue = controllerInput[0];
                    this._lastGP2con = controllerInput[0];
                    if (controllerInput[0]) {
                        this._GP2conBackground.background =
                            this._backgroundHighlightClr;
                    } else {
                        this._GP2conBackground.background =
                            this._backgroundClr;
                    }
                }

                if (controllerInput[1] != this._lastGP2map) {
                    this._GP2map.nodeValue = controllerInput[1];
                    this._lastGP2map = controllerInput[1];
                    if (controllerInput[1]) {
                        this._GP2mapBackground.background =
                            this._backgroundHighlightClr;
                    } else {
                        this._GP2mapBackground.background =
                            this._backgroundClr;
                    }
                }

                //axes
                if (controllerInput[0]) {
                    //axes 0
                    if (controllerInput[3]) {
                        this._GP2leftHZ.nodeValue =
                            controllerInput[3].toPrecision(4);
                        this._GP2leftHZBackground.background =
                            this._backgroundHighlightClr;
                        this._updateGP2[3] = true;
                    } else if (this._updateGP2[3]) {
                        this._GP2leftHZBackground.background =
                            this._backgroundClr;
                        this._GP2leftHZ.nodeValue = "0.000";
                        this._updateGP2[3] = false;
                    }

                    //axes 1
                    if (controllerInput[4]) {
                        this._GP2leftVT.nodeValue =
                            controllerInput[4].toPrecision(4);
                        this._GP2leftVTBackground.background =
                            this._backgroundHighlightClr;
                        this._updateGP2[4] = true;
                    } else if (this._updateGP2[4]) {
                        this._GP2leftVTBackground.background =
                            this._backgroundClr;
                        this._GP2leftVT.nodeValue = "0.000";
                        this._updateGP2[4] = false;
                    }

                    //axes 2
                    if (controllerInput[5]) {
                        this._GP2rightHZ.nodeValue =
                            controllerInput[5].toPrecision(4);
                        this._GP2rightHZBackground.background =
                            this._backgroundHighlightClr;
                        this._updateGP2[5] = true;
                    } else if (this._updateGP2[5]) {
                        this._GP2rightHZBackground.background =
                            this._backgroundClr;
                        this._GP2rightHZ.nodeValue = "0.000";
                        this._updateGP2[5] = false;
                    }

                    //axes 3
                    if (controllerInput[6]) {
                        this._GP2rightVT.nodeValue =
                            controllerInput[6].toPrecision(4);
                        this._GP2rightVTBackground.background =
                            this._backgroundHighlightClr;
                        this._updateGP2[6] = true;
                    } else if (this._updateGP2[6]) {
                        this._GP2rightVTBackground.background =
                            this._backgroundClr;
                        this._GP2rightVT.nodeValue = "0.000";
                        this._updateGP2[6] = false;
                    }

                    //left trigger
                    if (controllerInput[13]) {
                        this._GP2LeftTrigger.nodeValue =
                            controllerInput[13].toPrecision(4);
                        this._GP2LeftTriggerBackground.background =
                            this._backgroundHighlightClr;
                        this._updateGP2[13] = true;
                    } else if (this._updateGP2[13]) {
                        this._GP2LeftTriggerBackground.background =
                            this._backgroundClr;
                        this._GP2LeftTrigger.nodeValue = "0.000";
                        this._updateGP2[13] = false;
                    }

                    //left trigger
                    if (controllerInput[14]) {
                        this._GP2RightTrigger.nodeValue =
                            controllerInput[14].toPrecision(4);
                        this._GP2RightTriggerBackground.background =
                            this._backgroundHighlightClr;
                        this._updateGP2[14] = true;
                    } else if (this._updateGP2[14]) {
                        this._GP2RightTriggerBackground.background =
                            this._backgroundClr;
                        this._GP2RightTrigger.nodeValue = "0.000";
                        this._updateGP2[14] = false;
                    }

                    //left bumper
                    if (controllerInput[11]) {
                        this._GP2LeftBumper.nodeValue =
                            controllerInput[11];
                        this._GP2LeftBumperBackground.background =
                            this._backgroundHighlightClr;
                        this._updateGP2[11] = true;
                    } else if (this._updateGP2[11]) {
                        this._GP2LeftBumperBackground.background =
                            this._backgroundClr;
                        this._GP2LeftBumper.nodeValue = "0";
                        this._updateGP2[11] = false;
                    }

                    //right bumper
                    if (controllerInput[12]) {
                        this._GP2RightBumper.nodeValue =
                            controllerInput[12];
                        this._GP2RightBumperBackground.background =
                            this._backgroundHighlightClr;
                        this._updateGP2[12] = true;
                    } else if (this._updateGP2[12]) {
                        this._GP2RightBumperBackground.background =
                            this._backgroundClr;
                        this._GP2RightBumper.nodeValue = "0";
                        this._updateGP2[12] = false;
                    }

                    //A BUTTON
                    if (controllerInput[7]) {
                        this._GP2a.nodeValue =
                            controllerInput[7];
                        this._GP2aBackground.background =
                            this._backgroundHighlightClr;
                        this._updateGP2[7] = true;
                    } else if (this._updateGP2[7]) {
                        this._GP2aBackground.background =
                            this._backgroundClr;
                        this._GP2a.nodeValue = "0";
                        this._updateGP2[7] = false;
                    }

                    //B BUTTON
                    if (controllerInput[8]) {
                        this._GP2b.nodeValue =
                            controllerInput[8];
                        this._GP2bBackground.background =
                            this._backgroundHighlightClr;
                        this._updateGP2[8] = true;
                    } else if (this._updateGP2[8]) {
                        this._GP2bBackground.background =
                            this._backgroundClr;
                        this._GP2b.nodeValue = "0";
                        this._updateGP2[8] = false;
                    }

                    //X BUTTON
                    if (controllerInput[9]) {
                        this._GP2x.nodeValue =
                            controllerInput[9];
                        this._GP2xBackground.background =
                            this._backgroundHighlightClr;
                        this._updateGP2[9] = true;
                    } else if (this._updateGP2[9]) {
                        this._GP2xBackground.background =
                            this._backgroundClr;
                        this._GP2x.nodeValue = "0";
                        this._updateGP2[9] = false;
                    }

                    //Y BUTTON
                    if (controllerInput[10]) {
                        this._GP2y.nodeValue =
                            controllerInput[10];
                        this._GP2yBackground.background =
                            this._backgroundHighlightClr;
                        this._updateGP2[10] = true;
                    } else if (this._updateGP2[10]) {
                        this._GP2yBackground.background =
                            this._backgroundClr;
                        this._GP2y.nodeValue = "0";
                        this._updateGP2[10] = false;
                    }

                    //DPAD U/D/L/R
                    if (controllerInput[19]) {
                        this._GP2dpUp.nodeValue =
                            controllerInput[19];
                        this._GP2dpUpBackground.background =
                            this._backgroundHighlightClr;
                        this._updateGP2[19] = true;
                    } else if (this._updateGP2[19]) {
                        this._GP2dpUpBackground.background =
                            this._backgroundClr;
                        this._GP2dpUp.nodeValue = "0";
                        this._updateGP2[19] = false;
                    }

                    if (controllerInput[20]) {
                        this._GP2dpDown.nodeValue =
                            controllerInput[20];
                        this._GP2dpDownBackground.background =
                            this._backgroundHighlightClr;
                        this._updateGP2[20] = true;
                    } else if (this._updateGP2[20]) {
                        this._GP2dpDownBackground.background =
                            this._backgroundClr;
                        this._GP2dpDown.nodeValue = "0";
                        this._updateGP2[20] = false;
                    }

                    if (controllerInput[21]) {
                        this._GP2dpLeft.nodeValue =
                            controllerInput[21];
                        this._GP2dpLeftBackground.background =
                            this._backgroundHighlightClr;
                        this._updateGP2[21] = true;
                    } else if (this._updateGP2[21]) {
                        this._GP2dpLeftBackground.background =
                            this._backgroundClr;
                        this._GP2dpLeft.nodeValue = "0";
                        this._updateGP2[21] = false;
                    }

                    if (controllerInput[22]) {
                        this._GP2dpRight.nodeValue =
                            controllerInput[22];
                        this._GP2dpRightBackground.background =
                            this._backgroundHighlightClr;
                        this._updateGP2[22] = true;
                    } else if (this._updateGP2[22]) {
                        this._GP2dpRightBackground.background =
                            this._backgroundClr;
                        this._GP2dpRight.nodeValue = "0";
                        this._updateGP2[22] = false;
                    }

                    //other misc buttons
                    if (controllerInput[16]) {
                        this._GP2start.nodeValue =
                            controllerInput[16];
                        this._GP2startBackground.background =
                            this._backgroundHighlightClr;
                        this._updateGP2[16] = true;
                    } else if (this._updateGP2[16]) {
                        this._GP2startBackground.background =
                            this._backgroundClr;
                        this._GP2start.nodeValue = "0";
                        this._updateGP2[16] = true;
                    }

                    if (controllerInput[15]) {
                        this._GP2back.nodeValue =
                            controllerInput[15];
                        this._GP2backBackground.background =
                            this._backgroundHighlightClr;
                        this._updateGP2[15] = true;
                    } else if (this._updateGP2[15]) {
                        this._GP2backBackground.background =
                            this._backgroundClr;
                        this._GP2back.nodeValue = "0";
                        this._updateGP2[15] = false;
                    }

                    if (controllerInput[17]) {
                        this._GP2leftStick.nodeValue =
                            controllerInput[17];
                        this._GP2leftStickBackground.background =
                            this._backgroundHighlightClr;
                        this._updateGP2[17] = true;
                    } else if (this._updateGP2[17]) {
                        this._GP2leftStickBackground.background =
                            this._backgroundClr;
                        this._GP2leftStick.nodeValue = "0";
                        this._updateGP2[17] = false;
                    }

                    if (controllerInput[18]) {
                        this._GP2rightStick.nodeValue =
                            controllerInput[18];
                        this._GP2rightStickBackground.background =
                            this._backgroundHighlightClr;
                        this._updateGP2[18] = true;
                    } else if (this._updateGP2[18]) {
                        this._GP2rightStickBackground.background =
                            this._backgroundClr;
                        this._GP2rightStick.nodeValue = "0";
                        this._updateGP2[18] = false;
                    }

                }
            }

            //-------------------------------------------------------------
            // per frame activities - counters, caps, and timing
            //-------------------------------------------------------------
            this._renderCounter++
            if (this._renderCounter == 4) { this._renderCounter = 0; }            

            if (DISPLAY_UPDATE_FRAME_EQUALIZATION) {
                //figure out the frame time, and then pad that up to a set 
                //threthold to avoid trashing 
                this._updateTime = performance.now() - t;

                //essentially a 'time balancer' so we only update the DOM
                //every other frame.  This keeps EDGE from leaking 
                //memory for some reasons that are very confusing
                var ct = performance.now();
                while (true) {
                    var n = performance.now()
                    if (n - ct > (this._timingThreshold - this._updateTime)) {
                        break;
                    }
                }
            }
        }

        //-----------------------------------------------------------------
        //if we are debugging, and threading, then draw the bins / collision
        //stuff here from last known good data from threading pool 
        //(set via the setDebugBins() method)  //this pipeline not used
        //during the 'regular' non-threaded model
        //-----------------------------------------------------------------
        //draw all 2D debugging information - 
        this._debugOverlayContext.clearRect(0, 0,
            GAME_WINDOW.GAME_WINDOW_WIDTH, GAME_WINDOW.GAME_WINDOW_HEIGHT);

            
        //draw overlay to the 2d canvas if its on
        if (this._2dDebugVisibility) {
            this.drawDebugBox(0, 0,
                GAME_WINDOW.GAME_WINDOW_WIDTH,
                GAME_WINDOW.GAME_WINDOW_HEIGHT-1,
                'rgba(255,255,255,0.0)', 'rgba(255,255,255,1.0)');
        }

        //-----------------------------------------------------------------
        if (this._2dDebugVisibility && USE_THREADING && DEBUG) {
            var binWidth = GAME_WINDOW.GAME_WINDOW_UNSCALED_WIDTH /
                COLLISION_BINS_X;
            var binHeight = GAME_WINDOW.GAME_WINDOW_UNSCALED_HEIGHT /
                COLLISION_BINS_Y;

            if (this._debugBins.length > 0) {
                for (var i = 0; i < COLLISION_BINS_X; i++) {
                    for (var j = 0; j < COLLISION_BINS_Y; j++) {

                        var b = this._debugBins[i][j];

                        var x = Math.round((b._x1 /
                                GAME_WINDOW.GAME_WINDOW_UNSCALED_WIDTH) * 255);
                        var y = Math.round((b._y1 /
                                GAME_WINDOW.GAME_WINDOW_UNSCALED_HEIGHT) * 255);

                        if (b._gOs[0].length > 0) {

                            this.drawDebugBox(
                                 b._x1 * GAME_WINDOW.GAME_WINDOW_SCALE,
                                 b._y1 * GAME_WINDOW.GAME_WINDOW_SCALE,
                                 binWidth * GAME_WINDOW.GAME_WINDOW_SCALE,
                                 binHeight * GAME_WINDOW.GAME_WINDOW_SCALE,
                                 'rgba(' + x + ',0,' + y + ',0.8)',
                                 'rgba(255,255,255,0.5)');
                        } else {
                            this.drawDebugBox(
                                b._x1 * GAME_WINDOW.GAME_WINDOW_SCALE,
                                b._y1 * GAME_WINDOW.GAME_WINDOW_SCALE,
                                binWidth * GAME_WINDOW.GAME_WINDOW_SCALE,
                                binHeight * GAME_WINDOW.GAME_WINDOW_SCALE,
                                'rgba(' + x + ',0,' + y + ',0.2)',
                                'rgba(255,255,255,0.5)');
                        }
                    }
                }
            }
        }
        this._debugOverlayContext.fillStyle = 'rgba(0,0,0,0.0)';
        this._debugOverlayContext.lineWidth = 1;
    }
    //---------------------------------------------------------------------


    //------[  DEBUG DISPLAY::printDebugMessages]--------------------------
    // Purpose: print debug messages in the debug console, called from
    // overwritten console.log().  
    //---------------------------------------------------------------------
    printDebugMessage(msg) {
        this._debugConsoleMsgs.push(msg);

        if (this._debugText) {
            for (var i = 0; i < this._debugConsoleMsgs.length; i++) {
                if (i < this._debugConsoleLineHeight) {
                    this._debugText[i].nodeValue = this._debugConsoleMsgs[i];
                }
            }
        }

        if (this._debugConsoleMsgs.length >= this._debugConsoleLineHeight) {
            this._debugConsoleMsgs.splice(0, 1);
        }
    }
    //---------------------------------------------------------------------


    //------[  DEBUG DISPLAY::setDebugHandles]-----------------------------
    // Purpose: pick throught the HTML in debug_display div and get handles
    // to various fields and nodes to update them with the values from the
    // input manager and fps manager.  
    //---------------------------------------------------------------------
    setDebugHandles() {
        var e = 0; //temporary holder of various element handles

        //debug display toggle handles
        this._debugDisplay = document.getElementById("debug_display");
        this._visibility = false;

        //mouse debug handles
        this._mouseXvalue = document.getElementById("debugMouseX");
        e = document.createTextNode("");
        this._mouseXvalue.appendChild(e);
        this._mouseXvalue = e;
        this._mouseXBackground = document.getElementById(
            "debugMouseXBackground").style;

        this._mouseYvalue = document.getElementById("debugMouseY");
        e = document.createTextNode("");
        this._mouseYvalue.appendChild(e);
        this._mouseYvalue = e;
        this._mouseYBackground = document.getElementById(
            "debugMouseYBackground").style;

        this._mouseXlastValue = document.getElementById(
            "debugLastMouseX");
        e = document.createTextNode("");
        this._mouseXlastValue.appendChild(e);
        this._mouseXlastValue = e;
        this._mouseXlastValueBackground = document.getElementById(
            "debugLastMouseXBackground").style;

        this._mouseYlastValue = document.getElementById(
            "debugLastMouseY");
        e = document.createTextNode("");
        this._mouseYlastValue.appendChild(e);
        this._mouseYlastValue = e;
        this._mouseYlastValueBackground = document.getElementById(
            "debugLastMouseYBackground").style;

        this._mouseXrawValue = document.getElementById(
            "debugRawMouseX");
        e = document.createTextNode("");
        this._mouseXrawValue.appendChild(e);
        this._mouseXrawValue = e;

        this._mouseYrawValue = document.getElementById(
            "debugRawMouseY");
        e = document.createTextNode("");
        this._mouseYrawValue.appendChild(e);
        this._mouseYrawValue = e;

        this._mouseXrawValueBackground = document.getElementById(
            "debugRawMouseXBackground").style;
        this._mouseYrawValueBackground = document.getElementById(
            "debugRawMouseYBackground").style;

        this._mouseWheelDeltaValueBackground =
            document.getElementById(
            "debugMouseWheelTotalBackground").style;
        this._mouseWheelDeltaValue = document.getElementById(
            "debugMouseWheelTotal");
        e = document.createTextNode("");
        this._mouseWheelDeltaValue.appendChild(e);
        this._mouseWheelDeltaValue = e;
        
        this._mouseWheelEvent = document.getElementById(
            "debugMouseWheelEvent");
        e = document.createTextNode("");
        this._mouseWheelEvent.appendChild(e);
        this._mouseWheelEvent = e;
        this._mouseWheelEventBackground =
            document.getElementById(
            "debugMouseWheelEventBackground").style;

        this._mouseLeftValue = document.getElementById(
            "debugMouseLeft");
        e = document.createTextNode("");
        this._mouseLeftValue.appendChild(e);
        this._mouseLeftValue = e;
        this._mouseLeftValueBackground =
            document.getElementById(
            "debugMouseLeftBackground").style;

        this._mouseMiddleValue = document.getElementById(
            "debugMouseMiddle");
        e = document.createTextNode("");
        this._mouseMiddleValue.appendChild(e);
        this._mouseMiddleValue = e;
        this._mouseMiddleValueBackground =
            document.getElementById(
            "debugMouseMiddleBackground").style;

        this._mouseRightValue = document.getElementById(
            "debugMouseRight");
        e = document.createTextNode("");
        this._mouseRightValue.appendChild(e);
        this._mouseRightValue = e;
        this._mouseRightValueBackground =
            document.getElementById(
            "debugMouseRightBackground").style;

        this._mouseActiveValue = document.getElementById(
            "debugMouseActive");
        e = document.createTextNode("");
        this._mouseActiveValue.appendChild(e);
        this._mouseActiveValue = e;
        this._mouseActiveValueBackground =
            document.getElementById(
            "debugMouseActiveBackground").style;

        //keyboard debug handles
        this._spaceValue = document.getElementById('debugSpace');
        e = document.createTextNode("");
        this._spaceValue.appendChild(e);
        this._spaceValue = e;
        this._spaceValueBackground =
            document.getElementById('debugSpaceBackground').style;

        this._KBUp = document.getElementById('debugKBup');
        e = document.createTextNode("");
        this._KBUp.appendChild(e);
        this._KBUp = e;
        this._KBUpBackground =
            document.getElementById('debugKBupBackground').style;

        this._KBDown = document.getElementById('debugKBdown');
        e = document.createTextNode("");
        this._KBDown.appendChild(e);
        this._KBDown = e;
        this._KBDownBackground =
            document.getElementById('debugKBdownBackground').style;

        this._KBLeft = document.getElementById('debugKBleft');
        e = document.createTextNode("");
        this._KBLeft.appendChild(e);
        this._KBLeft = e;
        this._KBLeftBackground =
            document.getElementById('debugKBleftBackground').style;

        this._KBRight = document.getElementById('debugKBright');
        e = document.createTextNode("");
        this._KBRight.appendChild(e);
        this._KBRight = e;
        this._KBRightBackground =
            document.getElementById('debugKBrightBackground').style;

        //gamepad 1 handles
        this._GP1con = document.getElementById('debugGP1con');
        e = document.createTextNode("");
        this._GP1con.appendChild(e);
        this._GP1con = e;
        this._GP1conBackground =
            document.getElementById('debugGP1conBackground').style;

        this._GP1map = document.getElementById('debugGP1map');
        e = document.createTextNode("");
        this._GP1map.appendChild(e);
        this._GP1map = e;
        this._GP1map.nodeValue = "0";
        this._GP1mapBackground =
            document.getElementById('debugGP1mapBackground').style;

        this._GP1leftHZ = document.getElementById('debugGP1leftHZ');
        e = document.createTextNode("");
        this._GP1leftHZ.appendChild(e);
        this._GP1leftHZ = e;
        this._GP1leftHZ.nodeValue = "0.000";
        this._GP1leftHZBackground =
            document.getElementById('debugGP1leftHZbackground').style;

        this._GP1leftVT = document.getElementById('debugGP1leftVT');
        e = document.createTextNode("");
        this._GP1leftVT.appendChild(e);
        this._GP1leftVT = e;
        this._GP1leftVT.nodeValue = "0.000";
        this._GP1leftVTBackground =
            document.getElementById('debugGP1leftVTbackground').style;

        this._GP1rightHZ = document.getElementById('debugGP1rightHZ');
        e = document.createTextNode("");
        this._GP1rightHZ.appendChild(e);
        this._GP1rightHZ = e;
        this._GP1rightHZ.nodeValue = "0.000";
        this._GP1rightHZBackground =
            document.getElementById('debugGP1rightHZbackground').style;

        this._GP1rightVT = document.getElementById('debugGP1rightVT');
        e = document.createTextNode("");
        this._GP1rightVT.appendChild(e);
        this._GP1rightVT = e;
        this._GP1rightVT.nodeValue = "0.000";
        this._GP1rightVTBackground =
            document.getElementById('debugGP1rightVTbackground').style;

        this._GP1LeftTrigger = document.getElementById('debugGP1leftTrigger');
        e = document.createTextNode("");
        this._GP1LeftTrigger.appendChild(e);
        this._GP1LeftTrigger = e;
        this._GP1LeftTrigger.nodeValue = "0.000";
        this._GP1LeftTriggerBackground =
            document.getElementById('debugGP1leftTriggerBackground').style;

        this._GP1RightTrigger = document.getElementById('debugGP1rightTrigger');
        e = document.createTextNode("");
        this._GP1RightTrigger.appendChild(e);
        this._GP1RightTrigger = e;
        this._GP1RightTrigger.nodeValue = "0.000";
        this._GP1RightTriggerBackground =
            document.getElementById('debugGP1rightTriggerBackground').style;

        this._GP1LeftBumper = document.getElementById('debugGP1leftBumper');
        e = document.createTextNode("");
        this._GP1LeftBumper.appendChild(e);
        this._GP1LeftBumper = e;
        this._GP1LeftBumper.nodeValue = "0";
        this._GP1LeftBumperBackground =
            document.getElementById('debugGP1leftBumperBackground').style;

        this._GP1RightBumper = document.getElementById('debugGP1rightBumper');
        e = document.createTextNode("");
        this._GP1RightBumper.appendChild(e);
        this._GP1RightBumper = e;
        this._GP1RightBumper.nodeValue = "0";
        this._GP1RightBumperBackground =
            document.getElementById('debugGP1rightBumperBackground').style;

        this._GP1a = document.getElementById('debugGP1a');
        e = document.createTextNode("");
        this._GP1a.appendChild(e);
        this._GP1a = e;
        this._GP1a.nodeValue = "0";
        this._GP1aBackground =
            document.getElementById('debugGP1aBackground').style;

        this._GP1b = document.getElementById('debugGP1b');
        e = document.createTextNode("");
        this._GP1b.appendChild(e);
        this._GP1b = e;
        this._GP1b.nodeValue = "0";
        this._GP1bBackground =
            document.getElementById('debugGP1bBackground').style;

        this._GP1x = document.getElementById('debugGP1x');
        e = document.createTextNode("");
        this._GP1x.appendChild(e);
        this._GP1x = e;
        this._GP1x.nodeValue = "0";
        this._GP1xBackground =
            document.getElementById('debugGP1xBackground').style;

        this._GP1y = document.getElementById('debugGP1y');
        e = document.createTextNode("");
        this._GP1y.appendChild(e);
        this._GP1y = e;
        this._GP1y.nodeValue = "0";
        this._GP1yBackground =
            document.getElementById('debugGP1yBackground').style;

        this._GP1dpUp = document.getElementById('debugGP1dpUp');
        e = document.createTextNode("");
        this._GP1dpUp.appendChild(e);
        this._GP1dpUp = e;
        this._GP1dpUp.nodeValue = "0";
        this._GP1dpUpBackground =
            document.getElementById('debugGP1dpUpBackground').style;

        this._GP1dpDown = document.getElementById('debugGP1dpDown');
        e = document.createTextNode("");
        this._GP1dpDown.appendChild(e);
        this._GP1dpDown = e;
        this._GP1dpDown.nodeValue = "0";
        this._GP1dpDownBackground =
            document.getElementById('debugGP1dpDownBackground').style;
        
        this._GP1dpLeft = document.getElementById('debugGP1dpLeft');
        e = document.createTextNode("");
        this._GP1dpLeft.appendChild(e);
        this._GP1dpLeft = e;
        this._GP1dpLeft.nodeValue = "0";
        this._GP1dpLeftBackground =
            document.getElementById('debugGP1dpLeftBackground').style;

        this._GP1dpRight = document.getElementById('debugGP1dpRight');
        e = document.createTextNode("");
        this._GP1dpRight.appendChild(e);
        this._GP1dpRight = e;
        this._GP1dpRight.nodeValue = "0";
        this._GP1dpRightBackground =
            document.getElementById('debugGP1dpRightBackground').style;

        this._GP1start = document.getElementById('debugGP1start');
        e = document.createTextNode("");
        this._GP1start.appendChild(e);
        this._GP1start = e;
        this._GP1start.nodeValue = "0";
        this._GP1startBackground =
            document.getElementById('debugGP1startBackground').style;

        this._GP1back = document.getElementById('debugGP1back');
        e = document.createTextNode("");
        this._GP1back.appendChild(e);
        this._GP1back = e;
        this._GP1back.nodeValue = "0";
        this._GP1backBackground =
            document.getElementById('debugGP1backBackground').style;

        this._GP1leftStick = document.getElementById('debugGP1leftStick');
        e = document.createTextNode("");
        this._GP1leftStick.appendChild(e);
        this._GP1leftStick = e;
        this._GP1leftStick.nodeValue = "0";
        this._GP1leftStickBackground =
            document.getElementById('debugGP1leftStickBackground').style;

        this._GP1rightStick = document.getElementById('debugGP1rightStick');
        e = document.createTextNode("");
        this._GP1rightStick.appendChild(e);
        this._GP1rightStick = e;
        this._GP1rightStick.nodeValue = "0";
        this._GP1rightStickBackground =
            document.getElementById('debugGP1rightStickBackground').style;

        //gamepad 2 handles
        this._GP2con = document.getElementById('debugGP2con');
        e = document.createTextNode("");
        this._GP2con.appendChild(e);
        this._GP2con = e;
        this._GP2conBackground =
            document.getElementById('debugGP2conBackground').style;

        this._GP2map = document.getElementById('debugGP2map');
        e = document.createTextNode("");
        this._GP2map.appendChild(e);
        this._GP2map = e;
        this._GP2map.nodeValue = "0";
        this._GP2mapBackground =
            document.getElementById('debugGP2mapBackground').style;

        this._GP2leftHZ = document.getElementById('debugGP2leftHZ');
        e = document.createTextNode("");
        this._GP2leftHZ.appendChild(e);
        this._GP2leftHZ = e;
        this._GP2leftHZ.nodeValue = "0.000";
        this._GP2leftHZBackground =
            document.getElementById('debugGP2leftHZbackground').style;

        this._GP2leftVT = document.getElementById('debugGP2leftVT');
        e = document.createTextNode("");
        this._GP2leftVT.appendChild(e);
        this._GP2leftVT = e;
        this._GP2leftVT.nodeValue = "0.000";
        this._GP2leftVTBackground =
            document.getElementById('debugGP2leftVTbackground').style;

        this._GP2rightHZ = document.getElementById('debugGP2rightHZ');
        e = document.createTextNode("");
        this._GP2rightHZ.appendChild(e);
        this._GP2rightHZ = e;
        this._GP2rightHZ.nodeValue = "0.000";
        this._GP2rightHZBackground =
            document.getElementById('debugGP2rightHZbackground').style;

        this._GP2rightVT = document.getElementById('debugGP2rightVT');
        e = document.createTextNode("");
        this._GP2rightVT.appendChild(e);
        this._GP2rightVT = e;
        this._GP2rightVT.nodeValue = "0.000";
        this._GP2rightVTBackground =
            document.getElementById('debugGP2rightVTbackground').style;

        this._GP2LeftTrigger = document.getElementById('debugGP2leftTrigger');
        e = document.createTextNode("");
        this._GP2LeftTrigger.appendChild(e);
        this._GP2LeftTrigger = e;
        this._GP2LeftTrigger.nodeValue = "0.000";
        this._GP2LeftTriggerBackground =
            document.getElementById('debugGP2leftTriggerBackground').style;

        this._GP2RightTrigger = document.getElementById('debugGP2rightTrigger');
        e = document.createTextNode("");
        this._GP2RightTrigger.appendChild(e);
        this._GP2RightTrigger = e;
        this._GP2RightTrigger.nodeValue = "0.000";
        this._GP2RightTriggerBackground =
            document.getElementById('debugGP2rightTriggerBackground').style;

        this._GP2LeftBumper = document.getElementById('debugGP2leftBumper');
        e = document.createTextNode("");
        this._GP2LeftBumper.appendChild(e);
        this._GP2LeftBumper = e;
        this._GP2LeftBumper.nodeValue = "0";
        this._GP2LeftBumperBackground =
            document.getElementById('debugGP2leftBumperBackground').style;

        this._GP2RightBumper = document.getElementById('debugGP2rightBumper');
        e = document.createTextNode("");
        this._GP2RightBumper.appendChild(e);
        this._GP2RightBumper = e;
        this._GP2RightBumper.nodeValue = "0";
        this._GP2RightBumperBackground =
            document.getElementById('debugGP2rightBumperBackground').style;

        this._GP2a = document.getElementById('debugGP2a');
        e = document.createTextNode("");
        this._GP2a.appendChild(e);
        this._GP2a = e;
        this._GP2a.nodeValue = "0";
        this._GP2aBackground =
            document.getElementById('debugGP2aBackground').style;

        this._GP2b = document.getElementById('debugGP2b');
        e = document.createTextNode("");
        this._GP2b.appendChild(e);
        this._GP2b = e;
        this._GP2b.nodeValue = "0";
        this._GP2bBackground =
            document.getElementById('debugGP2bBackground').style;

        this._GP2x = document.getElementById('debugGP2x');
        e = document.createTextNode("");
        this._GP2x.appendChild(e);
        this._GP2x = e;
        this._GP2x.nodeValue = "0";
        this._GP2xBackground =
            document.getElementById('debugGP2xBackground').style;

        this._GP2y = document.getElementById('debugGP2y');
        e = document.createTextNode("");
        this._GP2y.appendChild(e);
        this._GP2y = e;
        this._GP2y.nodeValue = "0";
        this._GP2yBackground =
            document.getElementById('debugGP2yBackground').style;

        this._GP2dpUp = document.getElementById('debugGP2dpUp');
        e = document.createTextNode("");
        this._GP2dpUp.appendChild(e);
        this._GP2dpUp = e;
        this._GP2dpUp.nodeValue = "0";
        this._GP2dpUpBackground =
            document.getElementById('debugGP2dpUpBackground').style;

        this._GP2dpDown = document.getElementById('debugGP2dpDown');
        e = document.createTextNode("");
        this._GP2dpDown.appendChild(e);
        this._GP2dpDown = e;
        this._GP2dpDown.nodeValue = "0";
        this._GP2dpDownBackground =
            document.getElementById('debugGP2dpDownBackground').style;

        this._GP2dpLeft = document.getElementById('debugGP2dpLeft');
        e = document.createTextNode("");
        this._GP2dpLeft.appendChild(e);
        this._GP2dpLeft = e;
        this._GP2dpLeft.nodeValue = "0";
        this._GP2dpLeftBackground =
            document.getElementById('debugGP2dpLeftBackground').style;

        this._GP2dpRight = document.getElementById('debugGP2dpRight');
        e = document.createTextNode("");
        this._GP2dpRight.appendChild(e);
        this._GP2dpRight = e;
        this._GP2dpRight.nodeValue = "0";
        this._GP2dpRightBackground =
            document.getElementById('debugGP2dpRightBackground').style;

        this._GP2start = document.getElementById('debugGP2start');
        e = document.createTextNode("");
        this._GP2start.appendChild(e);
        this._GP2start = e;
        this._GP2start.nodeValue = "0";
        this._GP2startBackground =
            document.getElementById('debugGP2startBackground').style;

        this._GP2back = document.getElementById('debugGP2back');
        e = document.createTextNode("");
        this._GP2back.appendChild(e);
        this._GP2back = e;
        this._GP2back.nodeValue = "0";
        this._GP2backBackground =
            document.getElementById('debugGP2backBackground').style;

        this._GP2leftStick = document.getElementById('debugGP2leftStick');
        e = document.createTextNode("");
        this._GP2leftStick.appendChild(e);
        this._GP2leftStick = e;
        this._GP2leftStick.nodeValue = "0";
        this._GP2leftStickBackground =
            document.getElementById('debugGP2leftStickBackground').style;

        this._GP2rightStick = document.getElementById('debugGP2rightStick');
        e = document.createTextNode("");
        this._GP2rightStick.appendChild(e);
        this._GP2rightStick = e;
        this._GP2rightStick.nodeValue = "0";
        this._GP2rightStickBackground =
            document.getElementById('debugGP2rightStickBackground').style;

        //graphics debugging handles
        this._versionDisplay = document.getElementById('graphicsVersion');
        e = document.createTextNode("");
        this._versionDisplay.appendChild(e);
        this._versionDisplay = e;

        this._vendorDisplay = document.getElementById('graphicsVendor');
        e = document.createTextNode("");
        this._vendorDisplay.appendChild(e);
        this._vendorDisplay = e;

        this._rendererDisplay = document.getElementById('graphicsRenderer');
        e = document.createTextNode("");
        this._rendererDisplay.appendChild(e);
        this._rendererDisplay = e;

        this._sLversionDisplay = document.getElementById('graphicsSLVersion');
        e = document.createTextNode("");
        this._sLversionDisplay.appendChild(e);
        this._sLversionDisplay = e;

        this._callsDisplay = document.getElementById('graphicsCalls');
        e = document.createTextNode("");
        this._callsDisplay.appendChild(e);
        this._callsDisplay = e;

        this._verticesDisplay = document.getElementById('graphicsVertices');
        e = document.createTextNode("");
        this._verticesDisplay.appendChild(e);
        this._verticesDisplay = e;

        this._facesDisplay = document.getElementById('graphicsFaces');
        e = document.createTextNode("");
        this._facesDisplay.appendChild(e);
        this._facesDisplay = e;

        this._pointsDisplay = document.getElementById('graphicsPoints');
        e = document.createTextNode("");
        this._pointsDisplay.appendChild(e);
        this._pointsDisplay = e;

        this._numTexturesDisplay = document.getElementById('graphicsNumTextures');
        e = document.createTextNode("");
        this._numTexturesDisplay.appendChild(e);
        this._numTexturesDisplay = e;

        this._numShadersDisplay = document.getElementById('graphicsNumShaders');
        e = document.createTextNode("");
        this._numShadersDisplay.appendChild(e);
        this._numShadersDisplay = e;

        //threading debug handles
        this._threadsEnabledDisplay = document.getElementById('debugThreadingEnabled');
        e = document.createTextNode("");
        this._threadsEnabledDisplay.appendChild(e);
        this._threadsEnabledDisplay = e;
        this._threadsEnabledDisplay.nodeValue = "false";
        this._threadsEnabledDisplayBackground =
            document.getElementById('debugThreadingEnabledBackground').style;

        this._threadsObjUpdate = document.getElementById('debugThreadingObjUpdate');
        e = document.createTextNode("");
        this._threadsObjUpdate.appendChild(e);
        this._threadsObjUpdate = e;
        this._threadsObjUpdate.nodeValue = "0.0000";
        this._threadsObjUpdateBackground =
            document.getElementById('debugThreadingObjUpdateBackground').style;
        this._threadsObjUpdateBackground.background = 'rgba(0,255,0,0.5)';
        
        this._threadsCollisionDetect = document.getElementById('debugThreadingCollisionDetect');
        e = document.createTextNode("");
        this._threadsCollisionDetect.appendChild(e);
        this._threadsCollisionDetect = e;
        this._threadsCollisionDetect.nodeValue = "0.0000";
        this._threadsCollisionDetectBackground =
            document.getElementById('debugThreadingCollisionDetectBackground').style;
        this._threadsCollisionDetectBackground.background = 'rgba(0,0,255,0.5)';

        this._threadsCallback = document.getElementById('debugThreadingCallback');
        e = document.createTextNode("");
        this._threadsCallback.appendChild(e);
        this._threadsCallback = e;
        this._threadsCallback.nodeValue = "0.0000";
        this._threadsCallbackBackground =
            document.getElementById('debugThreadingCallbackBackground').style;
        this._threadsCallbackBackground.background = 'rgba(255,0,200,0.5)';

        this._threadInputTimingUpdate = document.getElementById('debugThreadingInputProcessing');
        e = document.createTextNode("");
        this._threadInputTimingUpdate.appendChild(e);
        this._threadInputTimingUpdate = e;
        this._threadInputTimingUpdate.nodeValue = "0.0000";
        this._threadsCallbackBackground =
            document.getElementById('debugThreadingInputProcessingBackground').style;
        this._threadsCallbackBackground.background = 'rgba(255,0,0,0.5)';

    }
    //---------------------------------------------------------------------


    //---------------------------------------------------------------------
    drawDebugBox(x, y, w, h, clr, strokeClr) {
        //test stuff...
        this._debugOverlayContext.beginPath();
        this._debugOverlayContext.strokeStyle = strokeClr;
        this._debugOverlayContext.lineWidth = 1;
        this._debugOverlayContext.fillStyle = clr;
        this._debugOverlayContext.rect(x, y, w, h);
        this._debugOverlayContext.fill();
        this._debugOverlayContext.closePath();
        this._debugOverlayContext.stroke();
        
    }
    //---------------------------------------------------------------------


    //---------------------------------------------------------------------
    drawDebugCircle(x, y, r, clr, strokeClr) {
        this._debugOverlayContext.beginPath();
        this._debugOverlayContext.strokeStyle = strokeClr;
        this._debugOverlayContext.lineWidth = 1;
        this._debugOverlayContext.fillStyle = clr;
        this._debugOverlayContext.arc(x, y, r, 0, Math.PI*2);
        this._debugOverlayContext.fill();
        this._debugOverlayContext.closePath();
        this._debugOverlayContext.stroke();
    }
    //---------------------------------------------------------------------


    //---------------------------------------------------------------------
    drawDebugPoly(x, y, points, S, clr, strokeClr) {
        this._debugOverlayContext.beginPath();
        this._debugOverlayContext.strokeStyle = strokeClr;
        this._debugOverlayContext.lineWidth = 1;
        this._debugOverlayContext.fillStyle = clr;

        //NOTE: we account for game window scale here so that
        //we keep the rotated points in gameObj._adjustedPoints unscaled
        //by view (ie in world coords)
        for (var i = 0; i < points.length; i++) {
            if (i == 0) { // move to...
                this._debugOverlayContext.moveTo(x + points[i][0] * S, y + points[i][1] * S);
            } else {
                this._debugOverlayContext.lineTo(x + points[i][0] * S, y + points[i][1] * S);
            }
        }
        this._debugOverlayContext.fill();
        this._debugOverlayContext.closePath();
        this._debugOverlayContext.stroke();
    }
    //---------------------------------------------------------------------

    updateThreadInfoDisplay(t1, t2, t3, t4) {
        if (this._renderCounter == 0 || !DISPLAY_UPDATE_SLICING) {
            this._threadsObjUpdate.nodeValue = t1.toString().slice(0, 6); 
            this._threadsCollisionDetect.nodeValue = t2.toString().slice(0, 6);
            this._threadInputTimingUpdate.nodeValue = t3.toString().slice(0, 6);
            this._threadsCallback.nodeValue = t4.toString().slice(0, 6); 
        }
    }

    setDebugBins(binArray) {
        this._debugBins = binArray;
    }

    //------[  DEBUG DISPLAY::setKeyboardEvent ]---------------------------
    // Purpose: called by INPUT_MANAGER when w keyboard event occurs.  
    //---------------------------------------------------------------------
    setKeyboardEvent() { this._keyboardEventFrame = true; }
    //---------------------------------------------------------------------

    //------[  DEBUG DISPLAY::setMouseEvent ]------------------------------
    // Purpose: called by INPUT_MANAGER when a mouse event occurs.  
    //---------------------------------------------------------------------
    setMouseEvent() { this._mouseEventFrame = true; }
    //---------------------------------------------------------------------

    //------[  DEBUG DISPLAY::setMouseWheelEvent ]-------------------------
    // Purpose: called by INPUT_MANAGER when a mouse wheel event occurs  
    //---------------------------------------------------------------------
    setMouseWheelEvent() { this._mouseWheelEventFrame = true; }
}
//-------------------------------------------------------------------------



//-------------------------------------------------------------------------
// creation of global DEBUG_DISPLAY reference so it cam be used from main.  
// Must occur after class definition
//-------------------------------------------------------------------------
var DEBUG_DISPLAY = DEBUG_DISPLAY || new DebugDisplay();
//-------------------------------------------------------------------------


//-------------------------------------------------------------------------
// A quick override of console functions so we can show them in debug mode
//-------------------------------------------------------------------------
(function () {
    
    console.log = function (message) {
        var txt = "";
        if(arguments.length > 1) {
            for (var i = 1; i < arguments.length; i++) {
                txt = txt + " " + arguments[i].toString();
            }
        }
        DEBUG_DISPLAY.printDebugMessage(message + txt);
        if (ECHO_NORMAL_CONSOLE) {
            DEBUG_CONSOLE.apply(console, arguments);
        }
    };
    console.warn = function (message) {
        DEBUG_DISPLAY.printDebugMessage(message);
        if (ECHO_NORMAL_CONSOLE) {
            DEBUG_CONSOLE.apply(console, arguments);
        }
    }
    console.info = function (message) {
        DEBUG_DISPLAY.printDebugMessage(message);
        if (ECHO_NORMAL_CONSOLE) {
            DEBUG_CONSOLE.apply(console, arguments);
        }
    }
    console.trace = function (message) {
        DEBUG_DISPLAY.printDebugMessage(message);
        if (ECHO_NORMAL_CONSOLE) {
            DEBUG_CONSOLE.apply(console, arguments);
        }
    }
})();
//-------------------------------------------------------------------------