﻿"use strict";
//-------[ FPS COUNTER ]---------------------------------------------------
/*  
    PURPOSE: Class for a FPS_COUNTER that keeps track of FPS, application
    uptime, peak and similar values for debug display.

     //-----[ TIME SLICE GRAPHCS ]-------------------------------------------
    NOTES: The top two graphs of the display indicate the amount of time (ms)
    used by the user-defined code in the main execution thread.  From within
    the main application loop (ie the one that is called from getAnimationFrame)
    you can create a timeSlice by calling FPS_COUNTER.setTimeSlice(clr) where
    clr is a string value for the color style you want the slice.  That slice
    will be the active slice until setTimeSlice is called again.  Any code
    executed between these calls is the amount of time taken by the slice.

    The graph on the left maps all slices to a total of 1 on the vertical axis.
    So the size of a slice relative to its peers indicates the amount of 
    time taken.  The graph on the right maps the time taken by the slice 
    relative to the total time taken per for the last frame, inclusive of both 
    the user portion and the browser.  At 60FPS, this would ideally be 
    16.667ms, and so on.  Note that very small time slices (less than 0.5ms)
    may not be visible on the graph on the right, if their height maps to
    less than a single pixel.  Note also that slices may not be visible on
    the left if there is another slice that dominates the percentage (ie 99%
    vs 1%). 
*/
//-------------------------------------------------------------------------
var FPS_COLOR = 'rgba(50,200,255,0.45)'; //color of FPS graph fill
var FPS_COLOR_STROKE = 'rgba(50,200,255, 1.0)'; //color of FPS graph line
//-------------------------------------------------------------------------


//-------------------------------------------------------------------------
class FPSCounter {
    constructor() {

        //pattern to ensure singleton of GAME_WINDOW object
        //there can be only one...
        if (!FPS_COUNTER) {
            FPS_COUNTER = this;

        } else {
            console.log(
            "AAENGINE: ERROR: Attempt to create multiple fps counter." +
            "\nAAENGINE: ERROR: Only first instanciation will be used.");
            return FPS_COUNTER;
        }
    };

    //-------[ FPS_COUNTER:INIT ]------------------------------------------
    /*
        PURPOSE: Called at application startup
    */
    //---------------------------------------------------------------------
    init() {
        //-----[ RECORD TIME AND POPULATE VARS ]---------------------------
        //vars for determining fps
        this._timer = new Date(); //timer that is reused to determine time
        this._oldTime = new Date().getTime(); //the previous time
        this._totalFrameTime = 0; //total amount of time per frame (ms)
        this._fps = 60; //current fps
        this._fpsValues = []; //array of fps values from current to past
        this._numFPSdisplaySlices = 63; //number of slices to store for disp
        this._timeSliceVals = []; //time slice values from current to past
        this._timeSliceValsArray = []; //array of arrays of time slice vals
        this._totalFrameTimes = []; //array of total frame times current to past
        this._timeSliceTotals = []; //array of exec slice totals current to past
        this._timeSliceIndex = 0; //index of which slice is current
        this._peakFPS = 0; //peak fps value (init value is min)
        this._lowestFPS = 9999; //lowest fps value (init value is max)
        this._averageFPS = 0; //average fps value (calced each frame)
        this._averageFrameTime = 0.000; //average time spent per frame (ms)
        this._averageTimeSlice = 0.000; //average time spent in code exec (ms)
        this._timeDelay = 0; //timer shim to even out operations

        //fill initial arrays with default values
        for (var i = 0; i < this._numFPSdisplaySlices; i++) {
            this._fpsValues[i] = 0.0000;
            this._timeSliceValsArray[i] = [];
            this._totalFrameTimes[i] = 0;
            this._timeSliceTotals[i] = 0;
        }

        //canvas and context elements for text display
        this._fpsCanvas = document.getElementById('fpsDisplayCanvas');
        this._fpsContext = this._fpsCanvas.getContext('2d');
        this._timeSliceCanvas = document.getElementById('timeSliceDisplayCanvas');
        this._timeSliceContext = this._timeSliceCanvas.getContext('2d');
        this._timeSliceFrameCanvas = document.getElementById('sliceByTimeCanvas');
        this._timeSliceFrameContext = this._timeSliceFrameCanvas.getContext('2d');

        var e = 0; //temporary holder for elements while creating
        //DOM elements for text display
        this._fpsValue = document.getElementById('fpsValue');
        e = document.createTextNode("");
        this._fpsValue.appendChild(e);
        this._fpsValue = e;

        this._fpsPeakValue = document.getElementById('fpsPeakValue');
        e = document.createTextNode("");
        this._fpsPeakValue.appendChild(e);
        this._fpsPeakValue = e;

        this._fpsLowValue = document.getElementById('fpsLowValue');
        e = document.createTextNode("");
        this._fpsLowValue.appendChild(e);
        this._fpsLowValue = e;

        this._fpsTotalFramesValue =
            document.getElementById('fpsTotalFramesValue');
        e = document.createTextNode("");
        this._fpsTotalFramesValue.appendChild(e);
        this._fpsTotalFramesValue = e;

        this._fpsAverageValue =
            document.getElementById('fpsAvgValue');
        e = document.createTextNode("");
        this._fpsAverageValue.appendChild(e);
        this._fpsAverageValue = e;

        this._fpsTotalUptimeValue =
            document.getElementById('fpsTotalUpValue');
        e = document.createTextNode("");
        this._fpsTotalUptimeValue.appendChild(e);
        this._fpsTotalUptimeValue = e;

        this._timeLastFrameValue =
            document.getElementById('timeLastFrame');
        e = document.createTextNode("");
        this._timeLastFrameValue.appendChild(e);
        this._timeLastFrameValue = e;

        this._avgLastFrameValue =
            document.getElementById('avgLastFrame');
        e = document.createTextNode("");
        this._avgLastFrameValue.appendChild(e);
        this._avgLastFrameValue = e;

        this._sliceLastFrameValue =
            document.getElementById('sliceLastFrame');
        e = document.createTextNode("");
        this._sliceLastFrameValue.appendChild(e);
        this._sliceLastFrameValue = e;

        this._avgSliceLastFrameValue =
            document.getElementById('avgSliceLastFrame');
        e = document.createTextNode("");
        this._avgSliceLastFrameValue.appendChild(e);
        this._avgSliceLastFrameValue = e;

        this._slice0Value = document.getElementById('slice0');
        e = document.createTextNode("");
        this._slice0Value.appendChild(e);
        this._slice0Value = e;
        this._slice0Background =
            document.getElementById('slice0Background');

        this._slice1Value = document.getElementById('slice1');
        e = document.createTextNode("");
        this._slice1Value.appendChild(e);
        this._slice1Value = e;
        this._slice1Background =
            document.getElementById('slice1Background');

        this._slice2Value = document.getElementById('slice2');
        e = document.createTextNode("");
        this._slice2Value.appendChild(e);
        this._slice2Value = e;
        this._slice2Background =
            document.getElementById('slice2Background');

        this._slice3Value = document.getElementById('slice3');
        e = document.createTextNode("");
        this._slice3Value.appendChild(e);
        this._slice3Value = e;
        this._slice3Background =
            document.getElementById('slice3Background');

        this._slice4Value = document.getElementById('slice4');
        e = document.createTextNode("");
        this._slice4Value.appendChild(e);
        this._slice4Value = e;
        this._slice4Background =
            document.getElementById('slice4Background');

        this._slice0tile = document.getElementById('slice0tile');
        this._slice1tile = document.getElementById('slice1tile');
        this._slice2tile = document.getElementById('slice2tile');
        this._slice3tile = document.getElementById('slice3tile');
        this._slice4tile = document.getElementById('slice4tile');

        //timeSlice storage array and timeSliceColors storage array
        this._timeSlices = [];
        this._sliceColors = [];

        //vars for determining and displaying uptime
        this._frameCount = 0; this._totalSec = 0; this._totalMin = 0;
        this._padCharS = ""; this._padCharM = ""; this._padCharH = "";

        //set slice scale for rendering graphs
        this.setSliceScale();
    }
    //---------------------------------------------------------------------


    //-------[ FPS_COUNTER:UPDATE ]----------------------------------------
    /*
        PURPOSE: Called Per-Frame, calculates FPS, uptime, peak, low, etc.,
        updates all fields, updates all slice graphs, etc.
    */
    //---------------------------------------------------------------------
    update(curTime) {
        //-------[ END TIME SLICE CAPTURE PRIOR TO FRAME CALCS ]-----------
        this.setTimeSlice('rgba(0,0,0,0.0)');

        //-------[ CALCULATE FPS FROM LAST FRAME ]-------------------------
        this._totalFrameTime = curTime - this._oldTime;
        this._fps = 1000 / (this._totalFrameTime); 
        this._oldTime = curTime;
        
        this._fpsValues.push(this._fps); //push at last
        this._fpsValues.shift(); //delete at first

        this._totalFrameTimes.push(this._totalFrameTime);
        this._totalFrameTimes.shift();

        //check and store peak
        if (this._fps < 0 || this._fps > 1000) { }
        else {
            if (this._fps > this._peakFPS) { this._peakFPS = this._fps; }
            if (this._fps < this._lowestFPS) { this._lowestFPS = this._fps; }
        }

        //draw FPS into FPS display...
        this._fpsContext.clearRect(0, 0,
            this._fpsCanvas.width, this._fpsCanvas.height);

        this._fpsContext.fillStyle = FPS_COLOR;
        this._fpsContext.beginPath();
        this._fpsContext.moveTo(-1, 151);

        for (var i = 0; i < this._numFPSdisplaySlices; i++) {
            var y = 150 - ((this._fpsValues[i] / 60.00) * 120)
            this._fpsContext.lineTo((i * 5)-1, y);
        }

        //complete the rest of the shape so it fills below line
        this._fpsContext.lineTo(315, 151);
        this._fpsContext.closePath();
        this._fpsContext.fill();
        this._fpsContext.strokeStyle = FPS_COLOR_STROKE;
        this._fpsContext.lineWidth = 2;
        this._fpsContext.stroke();

        //draw a solid green bar at 60 FPS (desired threshold)
        this._fpsContext.beginPath();
        this._fpsContext.strokeStyle = 'rgba(0,255,0, 0.75)';
        this._fpsContext.moveTo(0, 149 - 120);
        this._fpsContext.lineTo(365, 149 - 120);
        this._fpsContext.lineWidth = 1;
        this._fpsContext.stroke();

        //draw a solid orange bar at 30 FPS (minimum threshold)
        this._fpsContext.beginPath();
        this._fpsContext.strokeStyle = 'rgba(255,128,0,0.75)';
        this._fpsContext.moveTo(0, 149 - 60);
        this._fpsContext.lineTo(365, 149 - 60);
        this._fpsContext.stroke();

        //draw a solid red bar at 30 FPS (minimum threshold)
        this._fpsContext.beginPath();
        this._fpsContext.strokeStyle = 'rgba(255,0,0,0.75)';
        this._fpsContext.moveTo(0, 149 - 54);
        this._fpsContext.lineTo(365, 149 - 54);
        this._fpsContext.stroke();

        //calculate total time every 10 frames
        if (this._frameCount % 10 == 0) {
           this._curDate = new Date();
           this._elapsed_ms = this._curDate - this._timer;
           this._seconds = Math.round(this._elapsed_ms / 1000);
           this._minutes = Math.floor(this._seconds / 60);
           this._hours = Math.floor(this._minutes / 60);

           //round things
           this._totalSec = this.trimSecondsMinutes(this._seconds);
           this._totalMin = this.trimSecondsMinutes(this._minutes);
           
           //print the time
           if (this._totalSec < 10) { this._padCharS = "0"; }
               else { this._padCharS = ""; }
           if (this._totalMin < 10) { this._padCharM = "0"; }
               else { this._padCharM = ""; }
           if (this._hours < 10) { this._padCharH = "0"; }
               else { this._padCharH = ""; }
           this._fpsTotalUptimeValue.nodeValue =
               this._padCharH + this._hours + ":" +
               this._padCharM + this._totalMin + ":" +
               this._padCharS + this._totalSec;
        }

        //increment frame count
        this._frameCount++;

        //calc and print average FPS
        this._averageFPS += (this._fps - this._averageFPS)
            / this._frameCount;

        //calculate all time slices relatie to one another
        this._timeSliceTotal = (this._timeSlices[
            this._timeSlices.length - 1] - this._timeSlices[0]);
        
        //small fudge to avoid division by 0
        if (this._timeSliceTotal == 0) {
            this._timeSliceTotal = 0.00000000001;
        }

        //set the time slices and calculate cumulative slice totals
        for (var i = 0; i < this._timeSlices.length - 1; i++) {
            this._timeSlices[i] = (this._timeSlices[i + 1] -
                this._timeSlices[i]) / this._timeSliceTotal;
            this._timeSliceVals[i] = (this._timeSliceVals[i + 1] -
                this._timeSliceVals[i]);
            if (this._timeSliceVals[i] <= 0) {
                this._timeSliceVals[i] = 0.00001
            }
            if (i > 0) {
                this._timeSlices[i] += this._timeSlices[i - 1];
            }
        }

        //cycle arrays (push and slice)
        this._timeSliceValsArray.push(this._timeSlices.slice(0) );
        this._timeSliceValsArray.shift();
        this._timeSliceTotals.push(this._timeSliceTotal);
        this._timeSliceTotals.shift();
        
        //update timeSlices for graphing
        //clear the existing canvas elements
        this._timeSliceContext.clearRect(0, 0,
            this._timeSliceCanvas.width, this._timeSliceCanvas.height);

        //console.log(this._timeSliceFrameCanvas.height);
        this._timeSliceFrameContext.clearRect(0, 0,
            this._timeSliceFrameCanvas.width,
            this._timeSliceFrameCanvas.height);

        //-----------------------------------------------------------------
        //loop through existing time slice vals (array of arrays)
        //-----------------------------------------------------------------
        for (var y = 0; y <= this._timeSliceValsArray[0].length; y++) {
            if (y != this._timeSliceValsArray[0].length) {

                //set fill style for time slice graphs
                this._timeSliceContext.fillStyle = this._sliceColors[y];
                this._timeSliceFrameContext.fillStyle = this._sliceColors[y];
                this._timeSliceContext.strokeStyle = 'rgba(0,0,0,0.5)';
                this._timeSliceContext.beginPath();

                this._timeSliceFrameContext.strokeStyle = 'rgba(0,0,0,0.0)';
                this._timeSliceFrameContext.lineWidth = 0;
                this._timeSliceFrameContext.beginPath();

                //if we are are at first slice, set default start position
                if (y == 0) {
                    this._timeSliceContext.moveTo(0,
                        this._timeSliceCanvas.height);
                    this._timeSliceFrameContext.moveTo(
                        0, this._timeSliceFrameCanvas.height);
                    //for other slices, start at height of last slice
                } else {
                    this._timeSliceContext.moveTo(
                        0, this._timeSliceCanvas.height -
                        (this._timeSliceValsArray[0][y - 1]
                        * this._timeSliceCanvas.height * 0.99));
                    this._timeSliceFrameContext.lineTo(0,
                        this._timeSliceFrameCanvas.height -
                        (this._timeSliceValsArray[0][y - 1]
                        * this._timeSliceTotals[0] * 3));
                }
                //for last slice, start at right to draw back to left
                if (y == this._timeSliceValsArray[0].length - 1) {
                    this._timeSliceContext.moveTo(
                        this._timeSliceCanvas.width, 0);
                    this._timeSliceContext.fillStyle = 'rgba(0,0,0,0.0)';

                    this._timeSliceFrameContext.moveTo(
                        this._timeSliceFrameCanvas.width, 0);
                    this._timeSliceFrameContext.fillStyle = 'rgba(0,0,0,0.0)';

                }
                //draw the graph from left to right if not top slice
                if (y != this._timeSliceValsArray[0].length - 1) {
                    for (var x = 0; x < this._numFPSdisplaySlices; x++) {
                        this._timeSliceContext.lineTo(x * 5 - 1,
                            this._timeSliceCanvas.height -
                            (this._timeSliceValsArray[x][y] *
                            this._timeSliceCanvas.height * 0.99));
                        this._timeSliceFrameContext.lineTo(
                            x * 5 * this._sliceScale - 1,
                            this._timeSliceFrameCanvas.height -
                            (this._timeSliceValsArray[x][y] *
                            this._timeSliceTotals[x] * 3));

                    }
                }
                //if first slice, draw along bottom
                if (y == 0) {
                    this._timeSliceContext.lineTo(x * 5 - 1,
                        this._timeSliceCanvas.height);
                    this._timeSliceFrameContext.lineTo(
                        x * 5 * this._sliceScale - 1,
                        this._timeSliceFrameCanvas.height);
                    //if not first slice, draw back from right to left
                } else {

                    for (var x = this._numFPSdisplaySlices - 1; x >= 0; x--) {
                        this._timeSliceContext.lineTo(
                            x * 5 - 1, this._timeSliceCanvas.height -
                            (this._timeSliceValsArray[x][y - 1] *
                            this._timeSliceCanvas.height * 0.99));
                        this._timeSliceFrameContext.lineTo(
                            x * 5 * this._sliceScale - 1,
                            this._timeSliceFrameCanvas.height -
                            (this._timeSliceValsArray[x][y - 1] *
                            this._timeSliceTotals[x] * 3));
                    }
                }
                //if last slice, then draw along top of graph
                if (y == this._timeSliceValsArray[0].length - 1) {
                    this._timeSliceContext.lineTo(0, 0);
                    this._timeSliceFrameContext.lineTo(0, 0);
                }
                //close, fill, and stroke paths
                this._timeSliceContext.closePath();
                this._timeSliceContext.fill();
                this._timeSliceContext.stroke();

                this._timeSliceFrameContext.closePath();
                this._timeSliceFrameContext.fill();
                this._timeSliceFrameContext.stroke();

            } else {

                if (this._frameCount > this._numFPSdisplaySlices) {
                    //draw total frame time...
                    this._timeSliceFrameContext.beginPath();
                    this._timeSliceFrameContext.strokeStyle = FPS_COLOR_STROKE;
                    this._timeSliceFrameContext.fillStyle = FPS_COLOR;
                    this._timeSliceFrameContext.lineWidth = 2;
                    this._timeSliceFrameContext.moveTo(0, 0);

                    for (var x = 0; x < this._numFPSdisplaySlices; x++) {
                        this._timeSliceFrameContext.lineTo(
                            x * 5 * this._sliceScale - 1,
                            this._timeSliceFrameCanvas.height -
                            this._totalFrameTimes[x] * 3);
                    }
                    this._timeSliceFrameContext.stroke();

                    for (var x = this._numFPSdisplaySlices - 1; x >= 0; x--) {
                        this._timeSliceFrameContext.lineTo(
                            x * 5 * this._sliceScale - 1,
                            this._timeSliceFrameCanvas.height -
                            (this._timeSliceValsArray[x][y - 2] *
                            this._timeSliceTotals[x] * 3));
                    }

                    this._timeSliceFrameContext.closePath();
                    this._timeSliceFrameContext.fill();
                }
            }
        }
        //-----------------------------------------------------------------

        //draw a solid green bar at 60 FPS (desired threshold)
        this._timeSliceFrameContext.beginPath();
        this._timeSliceFrameContext.strokeStyle = 'rgba(0,255,0, 0.75)';
        this._timeSliceFrameContext.moveTo(0,
            this._timeSliceFrameCanvas.height - 16.6667 * 3);
        this._timeSliceFrameContext.lineTo(
            this._timeSliceFrameCanvas.width,
            this._timeSliceFrameCanvas.height - 16.6667 * 3);
        this._timeSliceFrameContext.lineWidth = 1;
        this._timeSliceFrameContext.stroke();

        //draw a solid green orange at 30 FPS (low threshold)
        this._timeSliceFrameContext.beginPath();
        this._timeSliceFrameContext.strokeStyle = 'rgba(255,128,0,0.75)';
        this._timeSliceFrameContext.moveTo(0,
            this._timeSliceFrameCanvas.height - 33.3334 * 3);
        this._timeSliceFrameContext.lineTo(
            this._timeSliceFrameCanvas.width,
            this._timeSliceFrameCanvas.height - 33.3334 * 3);
        this._timeSliceFrameContext.lineWidth = 1;
        this._timeSliceFrameContext.stroke();

        //draw a solid green orange at 24 FPS (minimum threshold)
        this._timeSliceFrameContext.beginPath();
        this._timeSliceFrameContext.strokeStyle = 'rgba(255,0,0,0.75)';
        this._timeSliceFrameContext.moveTo(0,
            this._timeSliceFrameCanvas.height - 41.6667 * 3);
        this._timeSliceFrameContext.lineTo(
            this._timeSliceFrameCanvas.width,
            this._timeSliceFrameCanvas.height - 41.6667 * 3);
        this._timeSliceFrameContext.lineWidth = 1;
        this._timeSliceFrameContext.stroke();

        //calculate averages for time per frame and time per exec slice
        if (this._totalFrameTime >= 0) {
            this._averageFrameTime += (this._totalFrameTime
                - this._averageFrameTime) / this._frameCount;
        }

        this._averageTimeSlice += (this._timeSliceTotal
             - this._averageTimeSlice) / this._frameCount;
        
        //-----------------------------------------------------------------
        //alternate update path (even frame)
        //-----------------------------------------------------------------
        if (FPS_COUNTER._frameCount % 2 == 0) {

            //update FPS display values
            this._fpsAverageValue.nodeValue = this._averageFPS.toPrecision(6);
            this._fpsTotalFramesValue.nodeValue = this._frameCount.toPrecision(6);
            this._fpsValue.nodeValue = this._fps.toPrecision(6);
            this._fpsPeakValue.nodeValue = this._peakFPS.toPrecision(6);
            this._fpsLowValue.nodeValue = this._lowestFPS.toPrecision(6);

            //update the debug display with these values
            this._timeLastFrameValue.nodeValue =
                this._totalFrameTime.toString().slice(0, 7);
            this._avgLastFrameValue.nodeValue =
                this._averageFrameTime.toString().slice(0, 7);
            this._sliceLastFrameValue.nodeValue =
                this._timeSliceTotal.toString().slice(0, 7);
            this._avgSliceLastFrameValue.nodeValue =
                this._averageTimeSlice.toString().slice(0, 7);

        //-----------------------------------------------------------------
        // update odd frame 
        //-----------------------------------------------------------------
        } else {
            if (this._sliceColors.length > 0) {
                this._slice0Value.nodeValue = (
                    this._timeSliceVals[0].toString().slice(0, 6));
                if (this._frameCount < 10) {
                    this._slice0Background.style.background = this._sliceColors[0];
                    this._slice0tile.style.background = this._sliceColors[0];
                }
            }
            if (this._sliceColors.length > 1) {
                this._slice1Value.nodeValue = (
                    this._timeSliceVals[1].toString().slice(0, 6));
                if (this._frameCount < 10) {
                    this._slice1Background.style.background = this._sliceColors[1];
                    this._slice1tile.style.background = this._sliceColors[1];
                }
            }
            if (this._sliceColors.length > 2) {
                this._slice2Value.nodeValue = (
                    this._timeSliceVals[2].toString().slice(0, 6));
                if (this._frameCount < 10) {
                    this._slice2Background.style.background = this._sliceColors[2];
                    this._slice2tile.style.background = this._sliceColors[2];
                }
            }
            if (this._sliceColors.length > 3) {
                this._slice3Value.nodeValue = (
                    this._timeSliceVals[3].toString().slice(0, 6));
                if (this._frameCount < 10) {
                    this._slice3Background.style.background = this._sliceColors[3];
                    this._slice3tile.style.background = this._sliceColors[3];
                }
            }
            if (this._sliceColors.length > 4) {
                this._slice4Value.nodeValue = (
                    this._timeSliceVals[4].toString().slice(0, 6));
                if (this._frameCount < 10) {
                    this._slice4Background.style.background = this._sliceColors[4];
                    this._slice4tile.style.background = this._sliceColors[4];
                }
            }
        }

        //reset any needed vars for next frame
        this._timeSliceIndex = 0;
        
    }
    //---------------------------------------------------------------------

    //---------------------------------------------------------------------
    // helper function for trimming and rounding time variables
    //---------------------------------------------------------------------
    trimSecondsMinutes(elapsed) {
        if (elapsed >= 60) { return this.trimSecondsMinutes(elapsed - 60); }
        return elapsed;
    }
    //---------------------------------------------------------------------

    //---------------------------------------------------------------------
    /*  
        FPS_COUNTER:setTimeSlice()  Purpose: called from main game loop,
        this function marks a particular moment as ending one debug
        slice and beginning another.  Also called at the start of fps
        monitoring to end all slices.  (FPS_Counter.update should be called
        last per frame)
    */
    //---------------------------------------------------------------------
    setTimeSlice(clr) {
        var t = performance.now();
        this._timeSlices[this._timeSliceIndex] = t;
        this._timeSliceVals[this._timeSliceIndex] = t;
        this._sliceColors[this._timeSliceIndex] = clr;
        this._timeSliceIndex++;
    }
    //---------------------------------------------------------------------

    //---------------------------------------------------------------------
    /*  
        FPS_COUNTER:setSliceScale()  Purpose: called on init or when window
        resizes to ensure scale between all the graphing functions
    */
    //---------------------------------------------------------------------
    setSliceScale() {
        if (this._timeSliceCanvas) {
            this._sliceScale = (parseFloat(window.getComputedStyle(
                 this._timeSliceCanvas, null).width) /
                 parseFloat(window.getComputedStyle(
                 this._timeSliceFrameCanvas, null).width));
        }
    }
}
//-------------------------------------------------------------------------


//-------------------------------------------------------------------------
// creation of global GAME_WINDOW reference so it cam be bound
// to the window.onresize event.  Must occur after class definition
//-------------------------------------------------------------------------
var FPS_COUNTER = FPS_COUNTER || new FPSCounter();
//-------------------------------------------------------------------------

